import requests
import json
import base64

def convert_screenshot_to_base64(screenshot_path):
    """
    Converts a screenshot file to a Base64-encoded string.

    Args:
        screenshot_path (str): Path to the screenshot file.

    Returns:
        str: Base64-encoded string of the image.
    """
    try:
        with open(screenshot_path, "rb") as image_file:
            base64_image = base64.b64encode(image_file.read()).decode("utf-8")
        return base64_image
    except FileNotFoundError:
        raise FileNotFoundError(f"The file at {screenshot_path} was not found.")
    except Exception as e:
        raise RuntimeError(f"An error occurred while converting the image: {e}")


def send_request_to_server(payload):
    # Server URL
    url = "http://localhost:5000/generate_action"
    
    payload["screenshot"] = convert_screenshot_to_base64(payload["screenshot_path"])
    del payload["screenshot_path"]

    # Set headers
    headers = {
        "Content-Type": "application/json"
    }
    
    print("sending payload", [f"{k}: {v}" for k, v in payload.items() if k != "screenshot"])

    try:
        # Send POST request
        response = requests.post(url, json=payload, headers=headers)
        
        # Check if request was successful
        response.raise_for_status()
        
        # Parse response
        result = response.json()
        
        print("Status:", result["status"])
        # print("\nParsed GUI:", json.dumps(result["parsed_gui"], indent=2))
        print("\nGenerated Action:", json.dumps(result["generated_action"], indent=2))
        
        return result

    except requests.exceptions.RequestException as e:
        print(f"Error making request: {e}")
        if hasattr(e.response, 'text'):
            print(f"Server response: {e.response.text}")
        return None

if __name__ == "__main__":
    
    # Example input data
    payload = {
        "uia_data": None,
        "screenshot_path": "/Users/yyyang/showlab/code/ootb_internal/computer_use_ootb_internal/honkai-star-rail-menu-resized.jpg",
        "query": "Help me to complete the mission 'Buds of Memories' in Star Rail",
        "action_history": "Open the menu interface.",
        "mode": "teach",
        
        # Optional parameters
        "user_id": "star_rail",
        "trace_id": "default_trace",
        "scale_factor": "1.0x",
        "os_name": "Windows",
        "date_time": "2024-01-01",
        "llm_model": "gpt-4"
    }
    send_request_to_server(payload) 


import os
import sys
import io
import logging
import time
import json
import platform
from typing import Callable
from enum import StrEnum
from collections.abc import Callable

import torch
from computer_use_ootb_internal.computer_use_demo.gui_agent.anthropic_agent import AnthropicActor
from computer_use_ootb_internal.computer_use_demo.executor.anthropic_executor import AnthropicExecutor
from computer_use_ootb_internal.computer_use_demo.executor.teachmode_executor import TeachModeExecutor
from computer_use_ootb_internal.computer_use_demo.gui_agent.actor.llm_actor import LLMActor
from computer_use_ootb_internal.computer_use_demo.gui_agent.planner.teach_mode_vlm_planner import TeachModeVLMPlanner, split_button_reference
from computer_use_ootb_internal.computer_use_demo.tools.screen_capture import get_screenshot
from computer_use_ootb_internal.computer_use_demo.gui_agent.gui_parser.simple_parser.gui_parser import parse_gui
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_vlm
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import encode_image

# Configure logging
logging.basicConfig(level=logging.INFO)

# Define constants
BETA_FLAG = "computer-use-2024-10-22"


class APIProvider(StrEnum):
    ANTHROPIC = "anthropic"
    BEDROCK = "bedrock"
    VERTEX = "vertex"
    OPENAI = "openai"
    QWEN = "qwen"


PROVIDER_TO_DEFAULT_MODEL_NAME = {
    APIProvider.ANTHROPIC: "claude-3-5-sonnet-20241022",
    APIProvider.BEDROCK: "anthropic.claude-3-5-sonnet-20241022-v2:0",
    APIProvider.VERTEX: "claude-3-5-sonnet-v2@20241022",
}


def sampling_loop_sync(
    model: str,
    provider: APIProvider | None,
    system_prompt_suffix: str,
    messages: list[dict],
    output_callback: Callable[[str, str], None],
    tool_output_callback: Callable[[str, str], None],
    api_response_callback: Callable[[dict], None],
    api_key: str,
    only_n_most_recent_images: int | None = None,
    max_tokens: int = 4096,
    selected_screen: int = 0,
    user_id: str = None,
    trace_id: str = None
):
    """
    Synchronous sampling loop for assistant/tool interactions.
    """
    # Set device for model
    if torch.cuda.is_available():
        device = torch.device("cuda")
    elif torch.backends.mps.is_available():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")
    print(f"Model initialized on device: {device}.")

    if model == "claude-3-5-sonnet-20241022":
        actor = AnthropicActor(
            model=model,
            provider=provider,
            system_prompt_suffix=system_prompt_suffix,
            api_key=api_key,
            api_response_callback=api_response_callback,
            max_tokens=max_tokens,
            only_n_most_recent_images=only_n_most_recent_images,
            selected_screen=selected_screen
        )
        executor = AnthropicExecutor(
            output_callback=output_callback,
            tool_output_callback=tool_output_callback,
            selected_screen=selected_screen
        )
    elif model.startswith("teach-mode"):
        executor = TeachModeExecutor(
            output_callback=output_callback,
            tool_output_callback=tool_output_callback,
            selected_screen=0
        )
        llm_mode = model.replace("teach-mode-", "")
        actor = LLMActor(llm_model=llm_mode, output_callback=output_callback, device=device, selected_screen=selected_screen)
        planner = TeachModeVLMPlanner(
            model=llm_mode,
            output_callback=output_callback,
            api_response_callback=api_response_callback,
            user_id=user_id,
            trace_id=trace_id,
            max_tokens=max_tokens,
            only_n_most_recent_images=only_n_most_recent_images,
            selected_screen=selected_screen,
            device=device,
            system_prompt_suffix=system_prompt_suffix
        )

        while True:
            print("Checking platform...")
            if platform.system() != "Windows":
                print("Teach mode supports only Windows.")
                # break
                # pass

            vlm_response = planner(messages=messages)
            next_action = json.loads(vlm_response).get("Next Action")
            next_action = split_button_reference(next_action)

            if not next_action:
                final_sc, final_sc_path = get_screenshot(selected_screen=selected_screen)
                output_callback(
                    f"No more actions. Task complete. Final State:\n<img src='data:image/png;base64,{encode_image(str(final_sc_path))}'>",
                    sender="bot"
                )
                break

            output_callback(f"Executing action: {next_action}", sender="bot")
            uia_data, screenshot_path = get_screenshot(selected_screen=selected_screen)
            gui = parse_gui(
                user_id=user_id,
                trace_id=trace_id,
                screenshot_path=screenshot_path,
                query=next_action,
                mode="teach"
            )
            actor_response = actor(next_action, gui, screenshot_path)
            messages.append({"role": "user", "content": actor_response["content"]})

            for message in executor(actor_response, messages):
                yield message

    else:
        raise ValueError("Invalid model selected.")


# Callbacks
def output_callback(response, sender):
    print(f"[{sender}]: {response}")


def api_response_callback(response, sender):
    print(f"[API]: {response}")


def tool_output_callback(tool_result, sender):
    print(f"[Tool]: {tool_result}")


# Environment setup
os.environ["OPENAI_API_KEY"] = "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"
os.environ["GEMINI_API_KEY"] = "AIzaSyDgjkdmiKIzbDk3QjXoviLMQDMtpYps7gM"
os.environ["OOTB_PATH"] = r"./computer_use_demo"
os.environ['GOOGLE_API_KEY'] = "AIzaSyA5v66c7lIhpAjpxiflKY4VU41AGHcENHM"

# Example usage
messages = [{"role": "user",
             "content": "Help me to complete the mission 'Buds of Memories' in Star Rail"}]

sampling_loop = sampling_loop_sync(
    model="teach-mode-gpt-4o-mini",
    provider=None,
    system_prompt_suffix="",
    messages=messages,
    output_callback=output_callback,
    tool_output_callback=tool_output_callback,
    api_response_callback=api_response_callback,
    api_key="your-api-key",
    selected_screen=0,
    user_id="star_rail",
    trace_id="default_trace"
)

for step in sampling_loop:
    print(step)
    time.sleep(2)


import os
from huggingface_hub import hf_hub_download, list_repo_files

# Specify the model repository and destination folder
model_repo = "showlab/ShowUI-2B"
destination_folder = "./showui-2b"

# Ensure the destination folder exists
os.makedirs(destination_folder, exist_ok=True)

# List all files in the repository
files = list_repo_files(repo_id=model_repo)

# Download each file to the destination folder
for file in files:
    file_path = hf_hub_download(repo_id=model_repo, filename=file, local_dir=destination_folder)
    print(f"Downloaded {file} to {file_path}")


import subprocess
import platform
import pyautogui
import asyncio
import base64
import os
import time
if platform.system() == "Darwin":
    import Quartz  # uncomment this line if you are on macOS
from enum import StrEnum
from pathlib import Path
from typing import Literal, TypedDict
from uuid import uuid4
from screeninfo import get_monitors

from PIL import ImageGrab, Image
from functools import partial

from anthropic.types.beta import BetaToolComputerUse20241022Param

from .base import BaseAnthropicTool, ToolError, ToolResult
from .run import run

OUTPUT_DIR = "./tmp/outputs"

TYPING_DELAY_MS = 12
TYPING_GROUP_SIZE = 50

Action = Literal[
    "key",
    "type",
    "mouse_move",
    "left_click",
    "left_click_drag",
    "right_click",
    "middle_click",
    "double_click",
    "screenshot",
    "cursor_position",
]


class Resolution(TypedDict):
    width: int
    height: int


MAX_SCALING_TARGETS: dict[str, Resolution] = {
    "XGA": Resolution(width=1024, height=768),  # 4:3
    "WXGA": Resolution(width=1280, height=800),  # 16:10
    "FWXGA": Resolution(width=1366, height=768),  # ~16:9
}


class ScalingSource(StrEnum):
    COMPUTER = "computer"
    API = "api"


class ComputerToolOptions(TypedDict):
    display_height_px: int
    display_width_px: int
    display_number: int | None


def chunks(s: str, chunk_size: int) -> list[str]:
    return [s[i : i + chunk_size] for i in range(0, len(s), chunk_size)]


def get_screen_details():
    screens = get_monitors()
    screen_details = []

    # Sort screens by x position to arrange from left to right
    sorted_screens = sorted(screens, key=lambda s: s.x)

    # Loop through sorted screens and assign positions
    primary_index = 0
    for i, screen in enumerate(sorted_screens):
        if i == 0:
            layout = "Left"
        elif i == len(sorted_screens) - 1:
            layout = "Right"
        else:
            layout = "Center"
        
        if screen.is_primary:
            position = "Primary" 
            primary_index = i
        else:
            position = "Secondary"
        screen_info = f"Screen {i + 1}: {screen.width}x{screen.height}, {layout}, {position}"
        screen_details.append(screen_info)

    return screen_details, primary_index


class ComputerTool(BaseAnthropicTool):
    """
    A tool that allows the agent to interact with the screen, keyboard, and mouse of the current computer.
    Adapted for Windows using 'pyautogui'.
    """

    name: Literal["computer"] = "computer"
    api_type: Literal["computer_20241022"] = "computer_20241022"
    width: int
    height: int
    display_num: int | None

    _screenshot_delay = 2.0
    _scaling_enabled = True

    @property
    def options(self) -> ComputerToolOptions:
        width, height = self.scale_coordinates(
            ScalingSource.COMPUTER, self.width, self.height
        )
        return {
            "display_width_px": width,
            "display_height_px": height,
            "display_number": self.display_num,
        }

    def to_params(self) -> BetaToolComputerUse20241022Param:
        return {"name": self.name, "type": self.api_type, **self.options}

    def __init__(self, selected_screen: int = 0, is_scaling: bool = True):
        super().__init__()

        # Get screen width and height using Windows command
        self.display_num = None
        self.offset_x = 0
        self.offset_y = 0
        self.selected_screen = selected_screen   
        self.is_scaling = is_scaling
        self.width, self.height = self.get_screen_size()     

        # Path to cliclick
        self.cliclick = "cliclick"
        self.key_conversion = {"Page_Down": "pagedown",
                               "Page_Up": "pageup",
                               "Super_L": "win",
                               "Escape": "esc"}
        
        system = platform.system()        # Detect platform
        if system == "Windows":
            screens = get_monitors()
            sorted_screens = sorted(screens, key=lambda s: s.x)
            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")
            screen = sorted_screens[self.selected_screen]
            bbox = (screen.x, screen.y, screen.x + screen.width, screen.y + screen.height)

        elif system == "Darwin":  # macOS
            max_displays = 32  # Maximum number of displays to handle
            active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]
            screens = []
            for display_id in active_displays:
                bounds = Quartz.CGDisplayBounds(display_id)
                screens.append({
                    'id': display_id, 'x': int(bounds.origin.x), 'y': int(bounds.origin.y),
                    'width': int(bounds.size.width), 'height': int(bounds.size.height),
                    'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
                })
            sorted_screens = sorted(screens, key=lambda s: s['x'])
            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")
            screen = sorted_screens[self.selected_screen]
            bbox = (screen['x'], screen['y'], screen['x'] + screen['width'], screen['y'] + screen['height'])
        else:  # Linux or other OS
            cmd = "xrandr | grep ' primary' | awk '{print $4}'"
            try:
                output = subprocess.check_output(cmd, shell=True).decode()
                resolution = output.strip().split()[0]
                width, height = map(int, resolution.split('x'))
                bbox = (0, 0, width, height)  # Assuming single primary screen for simplicity
            except subprocess.CalledProcessError:
                raise RuntimeError("Failed to get screen resolution on Linux.")
            
        self.offset_x = screen['x'] if system == "Darwin" else screen.x
        self.offset_y = screen['y'] if system == "Darwin" else screen.y
        self.bbox = bbox
        

    async def __call__(
        self,
        *,
        action: Action,
        text: str | None = None,
        coordinate: tuple[int, int] | None = None,
        **kwargs,
    ):
        print(f"action: {action}, text: {text}, coordinate: {coordinate}")
        if action in ("mouse_move", "left_click_drag"):
            if coordinate is None:
                raise ToolError(f"coordinate is required for {action}")
            if text is not None:
                raise ToolError(f"text is not accepted for {action}")
            if not isinstance(coordinate, (list, tuple)) or len(coordinate) != 2:
                raise ToolError(f"{coordinate} must be a tuple of length 2")
            # if not all(isinstance(i, int) and i >= 0 for i in coordinate):
            if not all(isinstance(i, int) for i in coordinate):
                raise ToolError(f"{coordinate} must be a tuple of non-negative ints")
            
            if self.is_scaling:
                x, y = self.scale_coordinates(
                    ScalingSource.API, coordinate[0], coordinate[1]
                )
            else:
                x, y = coordinate

            # print(f"scaled_coordinates: {x}, {y}")
            # print(f"offset: {self.offset_x}, {self.offset_y}")
            
            x += self.offset_x
            y += self.offset_y

            print(f"mouse move to {x}, {y}")
            
            if action == "mouse_move":
                pyautogui.moveTo(x, y)
                return ToolResult(output=f"Moved mouse to ({x}, {y})")
            elif action == "left_click_drag":
                current_x, current_y = pyautogui.position()
                pyautogui.dragTo(x, y, duration=0.5)  # Adjust duration as needed
                return ToolResult(output=f"Dragged mouse from ({current_x}, {current_y}) to ({x}, {y})")

        if action in ("key", "type"):
            if text is None:
                raise ToolError(f"text is required for {action}")
            if coordinate is not None:
                raise ToolError(f"coordinate is not accepted for {action}")
            if not isinstance(text, str):
                raise ToolError(output=f"{text} must be a string")

            if action == "key":
                # Handle key combinations
                keys = text.split('+')
                for key in keys:
                    key = self.key_conversion.get(key.strip(), key.strip())
                    key = key.lower()
                    pyautogui.keyDown(key)  # Press down each key
                for key in reversed(keys):
                    key = self.key_conversion.get(key.strip(), key.strip())
                    key = key.lower()
                    pyautogui.keyUp(key)    # Release each key in reverse order
                return ToolResult(output=f"Pressed keys: {text}")
            
            elif action == "type":
                pyautogui.typewrite(text, interval=TYPING_DELAY_MS / 1000)  # Convert ms to seconds
                screenshot_base64 = (await self.screenshot()).base64_image
                return ToolResult(output=text, base64_image=screenshot_base64)

        if action in (
            "left_click",
            "right_click",
            "double_click",
            "middle_click",
            "screenshot",
            "cursor_position",
            "left_press",
        ):
            if text is not None:
                raise ToolError(f"text is not accepted for {action}")
            if coordinate is not None:
                raise ToolError(f"coordinate is not accepted for {action}")

            if action == "screenshot":
                return await self.screenshot()
            elif action == "cursor_position":
                x, y = pyautogui.position()
                x, y = self.scale_coordinates(ScalingSource.COMPUTER, x, y)
                return ToolResult(output=f"X={x},Y={y}")
            else:
                if action == "left_click":
                    pyautogui.click()
                elif action == "right_click":
                    pyautogui.rightClick()
                elif action == "middle_click":
                    pyautogui.middleClick()
                elif action == "double_click":
                    pyautogui.doubleClick()
                elif action == "left_press":
                    pyautogui.mouseDown()
                    time.sleep(1)
                    pyautogui.mouseUp()
                return ToolResult(output=f"Performed {action}")
            
        raise ToolError(f"Invalid action: {action}")

    async def screenshot(self):
        
        import time
        time.sleep(1)
        
        """Take a screenshot of the current screen and return a ToolResult with the base64 encoded image."""
        output_dir = Path(OUTPUT_DIR)
        output_dir.mkdir(parents=True, exist_ok=True)
        path = output_dir / f"screenshot_{uuid4().hex}.png"

        ImageGrab.grab = partial(ImageGrab.grab, all_screens=True)

        # Detect platform
        system = platform.system()

        if system == "Windows":
            # Windows: Use screeninfo to get monitor details
            screens = get_monitors()

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s.x)

            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")

            screen = sorted_screens[self.selected_screen]
            bbox = (screen.x, screen.y, screen.x + screen.width, screen.y + screen.height)

        elif system == "Darwin":  # macOS
            # macOS: Use Quartz to get monitor details
            max_displays = 32  # Maximum number of displays to handle
            active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]

            # Get the display bounds (resolution) for each active display
            screens = []
            for display_id in active_displays:
                bounds = Quartz.CGDisplayBounds(display_id)
                screens.append({
                    'id': display_id,
                    'x': int(bounds.origin.x),
                    'y': int(bounds.origin.y),
                    'width': int(bounds.size.width),
                    'height': int(bounds.size.height),
                    'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
                })

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s['x'])

            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")

            screen = sorted_screens[self.selected_screen]
            bbox = (screen['x'], screen['y'], screen['x'] + screen['width'], screen['y'] + screen['height'])

        else:  # Linux or other OS
            cmd = "xrandr | grep ' primary' | awk '{print $4}'"
            try:
                output = subprocess.check_output(cmd, shell=True).decode()
                resolution = output.strip().split()[0]
                width, height = map(int, resolution.split('x'))
                bbox = (0, 0, width, height)  # Assuming single primary screen for simplicity
            except subprocess.CalledProcessError:
                raise RuntimeError("Failed to get screen resolution on Linux.")

        # Take screenshot using the bounding box
        screenshot = ImageGrab.grab(bbox=bbox)

        # Set offsets (for potential future use)
        self.offset_x = screen['x'] if system == "Darwin" else screen.x
        self.offset_y = screen['y'] if system == "Darwin" else screen.y

        print(f"target_dimension {self.target_dimension}")
        
        if not hasattr(self, 'target_dimension'):
            screenshot = self.padding_image(screenshot)
            self.target_dimension = MAX_SCALING_TARGETS["WXGA"]

        # Resize if target_dimensions are specified
        print(f"offset is {self.offset_x}, {self.offset_y}")
        print(f"target_dimension is {self.target_dimension}")
        screenshot = screenshot.resize((self.target_dimension["width"], self.target_dimension["height"]))

        # Save the screenshot
        screenshot.save(str(path))

        if path.exists():
            # Return a ToolResult instance instead of a dictionary
            return ToolResult(base64_image=base64.b64encode(path.read_bytes()).decode())
        
        raise ToolError(f"Failed to take screenshot: {path} does not exist.")

    def padding_image(self, screenshot):
        """Pad the screenshot to 16:10 aspect ratio, when the aspect ratio is not 16:10."""
        _, height = screenshot.size
        new_width = height * 16 // 10

        padding_image = Image.new("RGB", (new_width, height), (255, 255, 255))
        # padding to top left
        padding_image.paste(screenshot, (0, 0))
        return padding_image

    async def shell(self, command: str, take_screenshot=True) -> ToolResult:
        """Run a shell command and return the output, error, and optionally a screenshot."""
        _, stdout, stderr = await run(command)
        base64_image = None

        if take_screenshot:
            # delay to let things settle before taking a screenshot
            await asyncio.sleep(self._screenshot_delay)
            base64_image = (await self.screenshot()).base64_image

        return ToolResult(output=stdout, error=stderr, base64_image=base64_image)

    def scale_coordinates(self, source: ScalingSource, x: int, y: int):
        """Scale coordinates to a target maximum resolution."""
        if not self._scaling_enabled:
            return x, y
        ratio = self.width / self.height
        target_dimension = None

        for target_name, dimension in MAX_SCALING_TARGETS.items():
            # allow some error in the aspect ratio - not ratios are exactly 16:9
            if abs(dimension["width"] / dimension["height"] - ratio) < 0.02:
                if dimension["width"] < self.width:
                    target_dimension = dimension
                    self.target_dimension = target_dimension
                    # print(f"target_dimension: {target_dimension}")
                break

        if target_dimension is None:
            # TODO: currently we force the target to be WXGA (16:10), when it cannot find a match
            target_dimension = MAX_SCALING_TARGETS["WXGA"]
            self.target_dimension = MAX_SCALING_TARGETS["WXGA"]

        # should be less than 1
        x_scaling_factor = target_dimension["width"] / self.width
        y_scaling_factor = target_dimension["height"] / self.height
        if source == ScalingSource.API:
            if x > self.width or y > self.height:
                raise ToolError(f"Coordinates {x}, {y} are out of bounds")
            # scale up
            return round(x / x_scaling_factor), round(y / y_scaling_factor)
        # scale down
        return round(x * x_scaling_factor), round(y * y_scaling_factor)

    def get_screen_size(self):
        if platform.system() == "Windows":
            # Use screeninfo to get primary monitor on Windows
            screens = get_monitors()

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s.x)
            
            if self.selected_screen is None:
                primary_monitor = next((m for m in get_monitors() if m.is_primary), None)
                return primary_monitor.width, primary_monitor.height
            elif self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")
            else:
                screen = sorted_screens[self.selected_screen]
                return screen.width, screen.height

        elif platform.system() == "Darwin":
            # macOS part using Quartz to get screen information
            max_displays = 32  # Maximum number of displays to handle
            active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]

            # Get the display bounds (resolution) for each active display
            screens = []
            for display_id in active_displays:
                bounds = Quartz.CGDisplayBounds(display_id)
                screens.append({
                    'id': display_id,
                    'x': int(bounds.origin.x),
                    'y': int(bounds.origin.y),
                    'width': int(bounds.size.width),
                    'height': int(bounds.size.height),
                    'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
                })

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s['x'])

            if self.selected_screen is None:
                # Find the primary monitor
                primary_monitor = next((screen for screen in screens if screen['is_primary']), None)
                if primary_monitor:
                    return primary_monitor['width'], primary_monitor['height']
                else:
                    raise RuntimeError("No primary monitor found.")
            elif self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")
            else:
                # Return the resolution of the selected screen
                screen = sorted_screens[self.selected_screen]
                return screen['width'], screen['height']

        else:  # Linux or other OS
            cmd = "xrandr | grep ' primary' | awk '{print $4}'"
            try:
                output = subprocess.check_output(cmd, shell=True).decode()
                resolution = output.strip().split()[0]
                width, height = map(int, resolution.split('x'))
                return width, height
            except subprocess.CalledProcessError:
                raise RuntimeError("Failed to get screen resolution on Linux.")
    
    def get_mouse_position(self):
        # TODO: enhance this func
        from AppKit import NSEvent
        from Quartz import CGEventSourceCreate, kCGEventSourceStateCombinedSessionState

        loc = NSEvent.mouseLocation()
        # Adjust for different coordinate system
        return int(loc.x), int(self.height - loc.y)

    def map_keys(self, text: str):
        """Map text to cliclick key codes if necessary."""
        # For simplicity, return text as is
        # Implement mapping if special keys are needed
        return text



import os
import logging
import base64
from litellm import completion

# from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils import is_image_path, encode_image

def is_image_path(text):
    # Checking if the input text ends with typical image file extensions
    image_extensions = (".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".tif")
    if text.endswith(image_extensions):
        return True
    else:
        return False

def encode_image(image_path):
    """Encode image file to base64."""
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")
    
def run_litellm(messages: list, system: str, llm: str, max_tokens=256, temperature=0):
    """
    Support message format: 
    1. [
        {
            "role": "user",
            "content": ["What’s in this image?"]
        },
        {
            "role": "user",
            "content": ["E:/Workspace/computer_use_ootb_internal/examples/init_states/amazon.png"],
        }
    ],

    2. str: "What’s in this image?"
    """   

    final_messages = [{"role": "system", "content": system}]

    # image_url = "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
    if type(messages) == list:
        for item in messages:
            contents = []
            if isinstance(item, dict):
                for cnt in item["content"]:
                    if isinstance(cnt, str):
                        if is_image_path(cnt):
                            base64_image = encode_image(cnt)
                            content = {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"}}
                        # content = {"type": "image_url", "image_url": {"url": image_url}}
                        else:
                            content = {"type": "text", "text": cnt}
                    contents.append(content)
                    
                message = {"role": item["role"], "content": contents}
            else:  # str
                contents.append({"type": "text", "text": item})
                message = {"role": "user", "content": contents}
            
            final_messages.append(message)

    elif isinstance(messages, str):
        final_messages = [{"role": "user", "content": messages}]

    print(f"[litellm]-[{llm}] sending messages:", final_messages)
    

    response = completion(
        model = llm, 
        messages=final_messages,
        max_tokens=max_tokens,
        temperature=temperature,
    )

    try:
        text = response.choices[0].message.content
        token_usage = int(response.usage['total_tokens'])
        return text, token_usage
        
    # return error message if the response is not successful
    except Exception as e:
        print(f"Error in interleaved openAI: {e}. This may due to your invalid OPENAI_API_KEY. Please check the response: {response.json()} ")
        return response.json()


if __name__ == "__main__":
    system_prompt = "You are a helpful assistant."

    # Set the API key of model provider
    os.environ["OPENAI_API_KEY"] = "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"

    response, token_usage = run_litellm(
        messages=[
            {
                "role": "user",
                # "content": ["What’s in this image?"]
                'content': ["\nYou are using an Darwin device.\nYou are able to use a mouse and keyboard to interact with the computer based on the given task and screenshot.\nYou can only interact with the desktop GUI (no terminal or application menu access).\n\nYou may be given some history plan and actions, this is the response from the previous loop.\nYou should carefully consider your plan base on the task, screenshot, and history actions.\n\nYour available 'Next Action' only include:\n- ENTER: Press an enter key.\n- ESCAPE: Press an ESCAPE key.\n- INPUT: Input a string of text.\n- CLICK: Describe the ui element to be clicked.\n- HOVER: Describe the ui element to be hovered.\n- SCROLL: Scroll the screen, you must specify up or down.\n- PRESS: Describe the ui element to be pressed.\n\n\nOutput format:\n```json\n{\n    'Thinking': str, # describe your thoughts on how to achieve the task, choose one action from available actions at a time.\n    'Next Action': str, 'action_type, action description' | 'None' # one action at a time, describe it in short and precisely. \n}\n```\n\nOne Example:\n```json\n{  \n    'Thinking': 'I need to search and navigate to amazon.com.',\n    'Next Action': 'CLICK \'Search Google button\' <button_reference:icon_0004.png>.'\n}\n```\n\nIMPORTANT NOTES:\n0. Carefully observe the screenshot and read history actions.\n1. You should only give a single action at a time. for example, INPUT text, and ENTER can\'t be in one Next Action.\n2. Attach the text to Next Action, if there is text or any description for the button. \n3. You can provide the reference of the button, such as '<button_reference:icon_0004.png>', if you can find it in the in-context example.\n4. You should not include other actions, such as keyboard shortcuts.\n5. When the task is completed, you should say 'Next Action': 'None' in the json field.\nNOTE: Reference the following action trajectory to do the task, when user ask you to do the similar task.    \nIN-CONTEXT EXAMPLE:\n1: To enter the \'Cavern of Corrosion: Path of Providence\' dungeon and complete the \'Echo of War - Inner Beast\'s Battlefield\' for rewards, I need to click on the entrance.| icon_0003.png\n2: To access the dungeon or related menu, I need to click on the icon_0004 icon.| icon_0004.png\n3: To scroll down the list of available dungeons or materials, I need to use the \'Alt - MouseWheelDown\' shortcut.\n4: The user likely intended to scroll down the list of options or navigate through the interface using the mouse wheel down.\n5: To locate the \'Echo of War - Inner Beast\'s Battlefield\' dungeon, I need to scroll down the list.\n6: To view more options or details in the list, the user needs to scroll down.\n7: To proceed with the \'Echo of War - Inner Beast\'s Battlefield\' dungeon, I need to select the 'Echo of War' section.\n8: To view and select the \'Echo of War - Inner Beast\'s Battlefield\' dungeon, I need to click on the 'Echo of War' section.\n9: To complete the dungeon and receive rewards, I need to teleport to the \'Inner Beast\'s Battlefield\'.| icon_0016.png\n10: To access the \'Echo of War - Inner Beast\'s Battlefield\' dungeon, I need to click the \'Teleport\' button.| icon_0017.png\n11: To start the dungeon and receive the rewards, I need to click the \'Challenge\' button.| icon_0018.png\n12: To start the \'Echo of War - Inner Beast\'s Battlefield\' dungeon and receive the rewards, I need to click the \'Challenge\' button.| icon_0019.png\n13: To interact with or select a specific character or option, I need to click on the designated area.\n14: To interact with a specific UI element or character, I need to use \'Alt - LClick\' at the given coordinates.\n15: To begin the \'Echo of War - Inner Beast\'s Battlefield\' dungeon, I need to click the 'Start Challenge' button.| icon_0022.png\n16: To begin the dungeon and complete the task to receive rewards, I need to click the \'Start Challenge\' button.| icon_0023.png\n17: To proceed with the game or adjust settings, I need to click on the UI control.\n18: To leave the completed dungeon and return to the main game area, I need to click the \'Exit\' button.| icon_0025.png\n"]
            },
            {
                "role": "user",
                "content": ["/Users/yyyang/showlab/code/ootb_internal/computer_use_ootb_internal/output.png"],
            }
        ],
        system=system_prompt,
        llm="gpt-4o-mini",
    )

    print(response, token_usage)


import ast
import json
import asyncio
from typing import Any, Dict, cast, List, Union
from collections.abc import Callable
import uuid
from anthropic.types.beta import (
    BetaContentBlock,
    BetaContentBlockParam,
    BetaImageBlockParam,
    BetaMessage,
    BetaMessageParam,
    BetaTextBlockParam,
    BetaToolResultBlockParam,
)
from anthropic.types import TextBlock
from anthropic.types.beta import BetaMessage, BetaTextBlock, BetaToolUseBlock
from computer_use_ootb_internal.computer_use_demo.tools import BashTool, ComputerTool, EditTool, ToolCollection, ToolResult
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_showui, colorful_text_vlm


class TeachmodeShowUIExecutor:
    def __init__(
        self, 
        output_callback: Callable[[BetaContentBlockParam], None], 
        tool_output_callback: Callable[[Any, str], None],
        selected_screen: int = 0
    ):
        self.output_callback = output_callback
        self.tool_output_callback = tool_output_callback
        self.selected_screen = selected_screen
        self.screen_bbox = self._get_screen_resolution()
        print("Screen BBox:", self.screen_bbox)
        
        self.tool_collection = ToolCollection(
            ComputerTool(selected_screen=selected_screen, is_scaling=False)
        )
        
        self.supported_action_type={
            # "showui_action": "anthropic_tool_action"
            "CLICK": 'key',  # TBD
            "INPUT": "key",
            "ENTER": "key",  # TBD
            "ESC": "key",
            "ESCAPE": "key",
            "PRESS":  "key",
            "KEY": "key",
        }

    def __call__(self, response: str, messages: list[BetaMessageParam]):
        # response is expected to be :
        # {'content': "{'action': 'CLICK', 'value': None, 'position': [0.83, 0.15]}, ...", 'role': 'assistant'}, 
        
        action_dict = self._format_actor_output(response)  # str -> dict
        
        actions = action_dict["content"]
        role = action_dict["role"]
        
        # Parse the actions from showui
        action_list = self._parse_showui_output(actions)
        print("Parsed Action List:", action_list)
        
        tool_result_content = None
        
        if action_list is not None and len(action_list) > 0:
                    
            for action in action_list:  # Execute the tool (adapting the code from anthropic_executor.py)
            
                tool_result_content: list[BetaToolResultBlockParam] = []
                
                self.output_callback(f"{colorful_text_showui}:\n{action}", sender="bot")
                print("Converted Action:", action)
                
                sim_content_block = BetaToolUseBlock(id=f'toolu_{uuid.uuid4()}',
                                        input={'action': action["action"], 'text': action["text"], 'coordinate': action["coordinate"]},
                                        name='computer', type='tool_use')
                
                # update messages
                new_message = {
                    "role": "assistant",
                    "content": cast(list[BetaContentBlockParam], [sim_content_block]),
                }
                # if new_message not in messages:
                #     messages.append(new_message)

                # Run the asynchronous tool execution in a synchronous context
                result = asyncio.run(self.tool_collection.run(
                    name=sim_content_block.name,
                    tool_input=cast(dict[str, Any], sim_content_block.input),
                ))
                
                tool_result_content.append(
                    _make_api_tool_result(result, sim_content_block.id)
                )
                # print(f"executor: tool_result_content: {tool_result_content}")
                self.tool_output_callback(result, sim_content_block.id)

                # Craft messages based on the content_block
                # Note: to display the messages in the gradio, you should organize the messages in the following way (user message, bot message)
                display_messages = _message_display_callback(messages)
                
                # Send the messages to the gradio
                for user_msg, bot_msg in display_messages:
                    yield [user_msg, bot_msg], tool_result_content
        
        return tool_result_content
    
    
    def _format_actor_output(self, action_output: str|dict) -> Dict[str, Any]:
        if type(action_output) == dict:
            return action_output
        else:
            try:
                action_output.replace("'", "\"")
                action_dict = ast.literal_eval(action_output)
                return action_dict
            except Exception as e:
                print(f"Error parsing action output: {e}")
                return None
    

    def _parse_showui_output(self, output_text: str | dict) -> Union[List[Dict[str, Any]], None]:
        try:
            # refine key: value pairs, mapping to the Anthropic's format
            refined_output = []
            
            action_item = output_text
            
            print("[_parse_showui_output] Action Item:", action_item)
            
            # sometime showui returns lower case action names
            action_item["action"] = action_item["action"].upper()
            
            # if "coordinate" in action_item:
            #     action_item["position"] = action_item["coordinate"]
            # if "text" in action_item:
            #     action_item["value"] = action_item["text"]
            
            if action_item["action"] not in self.supported_action_type:
                raise ValueError(f"Action {action_item['action']} not supported. Check the output from ShowUI: {output_text}")
                # continue
            
            elif action_item["action"] == "CLICK":  # 1. click -> mouse_move + left_click
                x, y = action_item["position"]

                # action_item["position"] = (int(x * (self.screen_bbox[2] - self.screen_bbox[0])),
                #                             int(y * (self.screen_bbox[3] - self.screen_bbox[1])))
                action_item["position"] = (int(x), int(y))
                
                refined_output.append({"action": "mouse_move", "text": None, "coordinate": tuple(action_item["position"])})
                refined_output.append({"action": "left_click", "text": None, "coordinate": None})
            
            elif action_item["action"] == "INPUT":  # 2. input -> type
                if "text" in action_item:
                    refined_output.append({"action": "type", "text": action_item["text"], "coordinate": None})
                elif "value" in action_item:
                    refined_output.append({"action": "type", "text": action_item["value"], "coordinate": None})
                else:
                    raise ValueError(f"Input action {action_item['action']} does not contain 'text' or 'value'.")
            
            elif action_item["action"] == "ENTER":  # 3. enter -> key, enter
                refined_output.append({"action": "key", "text": "Enter", "coordinate": None})
            
            elif action_item["action"] == "ESC" \
                or action_item["action"] == "ESCAPE" \
                or (action_item["action"] == "KEY" and action_item["value"] == "ESC"):  # 4. enter -> key, enter
                refined_output.append({"action": "key", "text": "Escape", "coordinate": None})
                
            elif action_item["action"] == "HOVER":  # 5. hover -> mouse_move
                x, y = action_item["position"]
                action_item["position"] = (int(x * (self.screen_bbox[2] - self.screen_bbox[0])),
                                        int(y * (self.screen_bbox[3] - self.screen_bbox[1])))
                refined_output.append({"action": "mouse_move", "text": None, "coordinate": tuple(action_item["position"])})
                
            elif action_item["action"] == "SCROLL":  # 6. scroll -> key: pagedown
                if action_item["value"] == "up":
                    refined_output.append({"action": "key", "text": "pageup", "coordinate": None})
                elif action_item["value"] == "down":
                    refined_output.append({"action": "key", "text": "pagedown", "coordinate": None})
                else:
                    raise ValueError(f"Scroll direction {action_item['value']} not supported.")

            elif action_item["action"] == "PRESS":  # 7. press
                x, y = action_item["position"]
                action_item["position"] = (int(x * (self.screen_bbox[2] - self.screen_bbox[0])),
                                        int(y * (self.screen_bbox[3] - self.screen_bbox[1])))
                refined_output.append({"action": "mouse_move", "text": None, "coordinate": tuple(action_item["position"])})
                refined_output.append({"action": "left_press", "text": None, "coordinate": None})

            return refined_output

        except Exception as e:
            print(f"Error parsing output: {e}")
            import pdb; pdb.set_trace()
            return None
        

    def _get_screen_resolution(self):
        from screeninfo import get_monitors
        import platform
        if platform.system() == "Darwin":
            import Quartz  # uncomment this line if you are on macOS
        import subprocess
            
        # Detect platform
        system = platform.system()

        if system == "Windows":
            # Windows: Use screeninfo to get monitor details
            screens = get_monitors()

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s.x)

            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")

            screen = sorted_screens[self.selected_screen]
            bbox = (screen.x, screen.y, screen.x + screen.width, screen.y + screen.height)

        elif system == "Darwin":  # macOS
            # macOS: Use Quartz to get monitor details
            max_displays = 32  # Maximum number of displays to handle
            active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]

            # Get the display bounds (resolution) for each active display
            screens = []
            for display_id in active_displays:
                bounds = Quartz.CGDisplayBounds(display_id)
                screens.append({
                    'id': display_id,
                    'x': int(bounds.origin.x),
                    'y': int(bounds.origin.y),
                    'width': int(bounds.size.width),
                    'height': int(bounds.size.height),
                    'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
                })

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s['x'])

            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")

            screen = sorted_screens[self.selected_screen]
            bbox = (screen['x'], screen['y'], screen['x'] + screen['width'], screen['y'] + screen['height'])

        else:  # Linux or other OS
            cmd = "xrandr | grep ' primary' | awk '{print $4}'"
            try:
                output = subprocess.check_output(cmd, shell=True).decode()
                resolution = output.strip().split()[0]
                width, height = map(int, resolution.split('x'))
                bbox = (0, 0, width, height)  # Assuming single primary screen for simplicity
            except subprocess.CalledProcessError:
                raise RuntimeError("Failed to get screen resolution on Linux.")
        
        return bbox



def _message_display_callback(messages):
    display_messages = []
    for msg in messages:
        try:
            if isinstance(msg["content"][0], TextBlock):
                display_messages.append((msg["content"][0].text, None))  # User message
            elif isinstance(msg["content"][0], BetaTextBlock):
                display_messages.append((None, msg["content"][0].text))  # Bot message
            elif isinstance(msg["content"][0], BetaToolUseBlock):
                display_messages.append((None, f"Tool Use: {msg['content'][0].name}\nInput: {msg['content'][0].input}"))  # Bot message
            elif isinstance(msg["content"][0], Dict) and msg["content"][0]["content"][-1]["type"] == "image":
                display_messages.append((None, f'<img src="data:image/png;base64,{msg["content"][0]["content"][-1]["source"]["data"]}">'))  # Bot message
            else:
                pass
                # print(msg["content"][0])
        except Exception as e:
            print("error", e)
            pass
    return display_messages


def _make_api_tool_result(
    result: ToolResult, tool_use_id: str
) -> BetaToolResultBlockParam:
    """Convert an agent ToolResult to an API ToolResultBlockParam."""
    tool_result_content: list[BetaTextBlockParam | BetaImageBlockParam] | str = []
    is_error = False
    if result.error:
        is_error = True
        tool_result_content = _maybe_prepend_system_tool_result(result, result.error)
    else:
        if result.output:
            tool_result_content.append(
                {
                    "type": "text",
                    "text": _maybe_prepend_system_tool_result(result, result.output),
                }
            )
        if result.base64_image:
            tool_result_content.append(
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/png",
                        "data": result.base64_image,
                    },
                }
            )
    return {
        "type": "tool_result",
        "content": tool_result_content,
        "tool_use_id": tool_use_id,
        "is_error": is_error,
    }


def _maybe_prepend_system_tool_result(result: ToolResult, result_text: str):
    if result.system:
        result_text = f"<system>{result.system}</system>\n{result_text}"
    return result_text



# Testing main function
if __name__ == "__main__":
    def output_callback(content_block):
        # print("Output Callback:", content_block)
        pass

    def tool_output_callback(result, action):
        print("[showui_executor] Tool Output Callback:", result, action)
        pass

    # Instantiate the executor
    executor = ShowUIExecutor(
        output_callback=output_callback,
        tool_output_callback=tool_output_callback,
        selected_screen=0
    )

    # test inputs
    response_content = "{'content': \"{'action': 'CLICK', 'value': None, 'position': [0.49, 0.18]}\", 'role': 'assistant'}"
    # response_content = {'content': "{'action': 'CLICK', 'value': None, 'position': [0.49, 0.39]}", 'role': 'assistant'}
    # response_content = "{'content': \"{'action': 'CLICK', 'value': None, 'position': [0.49, 0.42]}, {'action': 'INPUT', 'value': 'weather for New York city', 'position': [0.49, 0.42]}, {'action': 'ENTER', 'value': None, 'position': None}\", 'role': 'assistant'}"

    # Initialize messages
    messages = []

    # Call the executor
    print("Testing ShowUIExecutor with response content:", response_content)
    for message, tool_result_content in executor(response_content, messages):
        print("Message:", message)
        print("Tool Result Content:", tool_result_content)

    # Display final messages
    print("\nFinal messages:")
    for msg in messages:
        print(msg)



import json
import uiautomation as auto
import time
import subprocess
import pygetwindow as gw
from pywinauto import Desktop, Application
import os
import re
import datetime
import time
import win32gui
import win32con
import win32api
from pywinauto.application import WindowSpecification
from pywinauto.findwindows import find_windows
from typing import List, Tuple, Dict
from PIL import Image, ImageDraw, ImageFont
from screeninfo import get_monitors


class Rectangle:
    def __init__(self, left: int, top: int, right: int, bottom: int):
        self.left = left
        self.top = top
        self.right = right
        self.bottom = bottom

    def intersects(self, other: 'Rectangle') -> bool:
        return not (self.right < other.left or
                   self.left > other.right or
                   self.bottom < other.top or
                   self.top > other.bottom)

    def get_intersection(self, other: 'Rectangle') -> 'Rectangle':
        left = max(self.left, other.left)
        top = max(self.top, other.top)
        right = min(self.right, other.right)
        bottom = min(self.bottom, other.bottom)
        if left < right and top < bottom:
            return Rectangle(left, top, right, bottom)
        return None

    def area(self) -> int:
        return (self.right - self.left) * (self.bottom - self.top)

    def is_within_monitor(self, monitor: 'Rectangle') -> bool:
        return self.intersects(monitor)


class VisibilityChecker:
    def __init__(self, selected_screen: int = 0):
        self.window_regions: List[Tuple[Rectangle, int, str]] = []
        self.monitor = self.get_monitor(selected_screen)

    def get_monitor(self, selected_screen: int = 0):
        screens = get_monitors()

        # Sort screens by x position to arrange from left to right
        sorted_screens = sorted(screens, key=lambda s: s.x)

        monitor = sorted_screens[selected_screen]

        return Rectangle(monitor.x, monitor.y, monitor.x + monitor.width, monitor.y + monitor.height)

    def add_window(self, rect: List[int], window_title: str = "") -> int:
        """添加窗口，记录其Z序和区域，返回分配的Z序"""
        rect_obj = Rectangle(*rect)
        if rect_obj.is_within_monitor(self.monitor):
            self.window_regions.append((rect_obj, window_title))

    def is_visible(self, rect: List[int], threshold: float = 0.3) -> bool:
        """
        检查元素是否可见
        Args:
            rect: 元素矩形区域 [left, top, right, bottom]
            parent_window_z_index: 父窗口的Z序
            threshold: 可见度阈值，默认为0.8（80%可见才算可见）
        """
        element_rect = Rectangle(*rect)
        
        # 检查是否在显示器范围内
        if not element_rect.is_within_monitor(self.monitor):
            return False

        element_area = element_rect.area()
        if element_area == 0:
            return False

        # 计算被其他窗口遮挡的面积
        covered_area = 0
        for window_rect, _ in self.window_regions:            
            intersection = element_rect.get_intersection(window_rect)
            if intersection:
                covered_area += intersection.area()

        # 计算可见比例
        visible_area = element_area - covered_area
        visibility_ratio = visible_area / element_area
        
        return visibility_ratio >= threshold

def get_window_z_order():
    def callback(hwnd, windows):
        if win32gui.IsWindowVisible(hwnd):
            windows.append(hwnd)
        return True
    
    windows = []
    win32gui.EnumWindows(callback, windows)
    return windows

class GUICapture:
    def __init__(self, cache_folder='.cache/', selected_screen: int = 0):
        self.task_id = self.get_current_time()
        self.cache_folder = os.path.join(cache_folder, self.task_id)
        os.makedirs(self.cache_folder, exist_ok=True)
        self.current_step = 0
        self.history = []
        self.visibility_checker = VisibilityChecker(selected_screen)
        
    def get_current_time(self):
        return datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    def is_element_visible(self, element) -> bool:
        try:
            if not element.is_visible():
                return False

            try:
                rect = element.rectangle()
                if rect.left < -32000 or rect.top < -32000 or rect.right > 32000 or rect.bottom > 32000:
                    return False
                if rect.right <= rect.left or rect.bottom <= rect.top:
                    return False
                
                rect_list = [rect.left, rect.top, rect.right, rect.bottom]
                return self.visibility_checker.is_visible(rect_list)
            except Exception as e:
                print(f"Error getting element rectangle: {e}")
                return False
                
        except Exception as e:
            print(f"Error checking visibility: {e}")
            return False

    def get_gui_meta_data(self):
        control_properties_list = ['friendly_class_name', 'texts', 'rectangle', 'automation_id', "value"]
        
        def get_window_controls(window_element):
            try:
                control_data = []
                try:
                    children = window_element.children()
                except Exception as e:
                    print(f"Error getting children: {e}")
                    return []
                
                for child in children:
                    try:                        
                        if not self.is_element_visible(child):
                            continue
                        
                        control_data.append({
                            'properties': get_control_properties(child, control_properties_list),
                            'children': get_window_controls(child)
                        })

                    except Exception as e:
                        print(f"Error processing child element: {e}")
                        continue
                        
                return control_data
            
            except Exception as e:
                print(f"Error getting controls: {e}")
                return []

        meta_data = {}
        desktop = Desktop(backend='uia')
        
        # 获取Z序排序的窗口句柄
        z_ordered_handles = get_window_z_order()
        windows = desktop.windows()
        handle_to_window = {win.handle: win for win in windows if win.is_visible()}
        
        # 处理任务栏
        try:
            print("Processing Taskbar")
            taskbar = desktop.window(class_name='Shell_TrayWnd')
            meta_data['Taskbar'] = get_window_controls(taskbar)
            rect = taskbar.rectangle()
            self.visibility_checker.add_window(
                [rect.left, rect.top, rect.right, rect.bottom],
                "Taskbar"
            )
        except Exception as e:
            print(f"Error getting taskbar: {e}")
            meta_data['Taskbar'] = []

        # 处理所有窗口
        for handle in z_ordered_handles:  # 从顶层窗口开始处理
            if handle in handle_to_window:
                window = handle_to_window[handle]
                try:
                    window_title = window.window_text()
                    if window_title and window_title != "Taskbar":
                        print(f"Processing window: {window_title}")
                        meta_data[window_title] = get_window_controls(window)

                        rect = window.rectangle()
                        self.visibility_checker.add_window(
                            [rect.left, rect.top, rect.right, rect.bottom],
                            window_title
                        )

                except Exception as e:
                    print(f"Error processing window: {e}")
                    continue

        return meta_data

    def capture_screenshot(self, save_path=None):
        # TODO: capture specific monitor
        if save_path:
            screenshot_path = save_path
        else:
            screenshot_path = os.path.join(self.cache_folder, f'screenshot-{self.current_step}.png')

        screenshot = auto.GetRootControl().ToBitmap()
        screenshot.ToFile(screenshot_path)
        return screenshot_path

    def capture(self):
        start = time.time()
        meta_data = self.get_gui_meta_data()
        screenshot_path = self.capture_screenshot()
        print(f"Time used: {time.time() - start}")
        return meta_data, screenshot_path

def get_control_properties(control, properties_list):
    prop_dict = {}
    for prop in properties_list:
        if prop == 'texts':
            continue

        if hasattr(control, prop):
            attr = getattr(control, prop)
            if callable(attr):
                try:
                    value = attr()
                    if prop == 'rectangle':
                        value = [value.left, value.top, value.right, value.bottom]

                    prop_dict[prop] = value
                except Exception:
                    continue
            else:
                prop_dict[prop] = attr

    # Get texts property
    if prop_dict['friendly_class_name'] in ['ComboBox']:
        prop_dict['texts'] = ['']
    else:
        attr = getattr(control, 'texts')
        value = attr()
        prop_dict['texts'] = value

    # Get value property
    # This is for file explorer, the texts do not contain the file name, but in value property
    if prop_dict['texts'] in [['名称'], ['修改日期'], ['大小'], ['类型'], ['Name'], ['Modified'], ['Size'], ['Type']]:
        try:
            pattern = control.element_info.element.GetCurrentPropertyValue(30045)  # UIA_ValueValuePropertyId
                
            if pattern:
                prop_dict['value'] = pattern
                if len(prop_dict['texts']) == 1:
                    prop_dict['texts'] = [prop_dict['value']]
            else:
                prop_dict['value'] = ''
        except Exception:
            prop_dict['value'] = ''

    return prop_dict

def visualize(gui, screenshot_path, if_show=True):
    ui_elements = []

    def extract_elements(node):
        if isinstance(node, list):
            for item in node:
                extract_elements(item)
        elif isinstance(node, dict):
            properties = node.get('properties', {})
            texts = properties.get('texts', [])
            rectangle = properties.get('rectangle', [])
            if rectangle and texts:
                name = texts
                ui_elements.append((name, rectangle))
            children = node.get('children', [])
            extract_elements(children)

    for window_name, panels in gui.items():
        extract_elements(panels)

    image = Image.open(screenshot_path)
    draw = ImageDraw.Draw(image)
    
    # 使用系统默认字体，通常支持中文
    try:
        # 优先尝试微软雅黑
        font = ImageFont.truetype("msyh.ttc", 12)
    except:
        try:
            # 备选宋体
            font = ImageFont.truetype("simsun.ttc", 12)
        except:
            # 如果都失败了，使用系统默认
            font = ImageFont.load_default()

    for element in ui_elements:
        try:
            name, rectangle = element
            if isinstance(name, list):
                if len(name) == 1:
                    name = name[0]  # 直接使用原始文本，不需要编码解码
                    if len(name) > 100:
                        name = name[:100] + "..."
                else:
                    print("name is list, but len(name) != 1: ", name)
                    continue
            else:
                print("name is not list: ", name)
                continue
        except Exception as e:
            print(f"Error processing element: {e}")
            continue
        
        draw.rectangle(rectangle, outline="red")
        if if_show:
            try:
                draw.text((rectangle[0], rectangle[1]), name, fill="red", font=font)
            except Exception as e:
                print(f"Error drawing text: {e}")

    return image

def get_screenshot(selected_screen: int = 0):
    gui = GUICapture(selected_screen=selected_screen)
    meta_data, screenshot_path = gui.capture()
    return meta_data, screenshot_path

def get_uia_data(selected_screen: int = 0):
    gui = GUICapture(selected_screen=selected_screen)
    uia_data, _ = gui.capture()
    return uia_data

if __name__ == '__main__':
    # 使用主显示器
    gui = GUICapture()
    # 或指定显示器
    # monitor_handle = win32api.MonitorFromPoint((0,0))
    # gui = GUICapture(monitor_handle=monitor_handle)
    
    meta_data, screenshot_path = gui.capture()
    print(meta_data)
    print(meta_data)


import os
import sys
import io
import time
import json
import platform
from typing import Callable
from collections.abc import Callable

from computer_use_ootb_internal.computer_use_demo.executor.teachmode_executor import TeachmodeShowUIExecutor
from computer_use_ootb_internal.computer_use_demo.gui_agent.gui_parser.simple_parser.gui_capture import get_screenshot
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import encode_image
from computer_use_ootb_internal.computer_use_demo.gui_agent.gui_parser.simple_parser.icon_detection.icon_detection import get_screen_resize_factor
from aws_request import send_request_to_server


def load_from_storage(file_name):
    with open(file_name, "r") as file:
        return json.load(file)

def simple_teachmode_sampling_loop(
    model: str,
    task: str,
    output_callback: Callable[[str, str], None],
    tool_output_callback: Callable[[str, str], None],
    api_response_callback: Callable[[dict], None],
    api_keys: dict = None,
    action_history: list[dict] = [],
    selected_screen: int = 0,
    user_id: str = None,
    trace_id: str = None,

):
    """
    Synchronous sampling loop for assistant/tool interactions.
    """
    
    print("Checking platform...")
    if platform.system() != "Windows":
        raise ValueError("Teach mode now only supports Windows.")

    
    executor = TeachmodeShowUIExecutor(
            output_callback=output_callback,
            tool_output_callback=tool_output_callback,
            selected_screen=selected_screen
        )
    
    step_count = 0

    if model.startswith("teach-mode"):
        
        model_name = model.replace("teach-mode-", "")
        
        while True:

            time.sleep(2)

            uia_meta, sc_path = get_screenshot(selected_screen=selected_screen)

            print("uia_meta:", uia_meta)

            payload = {
                "uia_data": uia_meta,
                "screenshot_path": sc_path,
                "query": task,
                "action_history": action_history,
                "mode": "teach",
                "user_id": user_id,
                "trace_id": trace_id,
                "scale_factor": get_screen_resize_factor(),  
                "os_name": platform.system(),
                "llm_model": "gpt-4o",
                "api_keys": api_keys
            }
            
            infer_server_response = send_request_to_server(payload)
            print("infer_server_response-", infer_server_response, "-infer_server_response")
            # import pdb; pdb.set_trace()
            
            next_plan = infer_server_response['generated_plan']

            try:
                next_action = json.loads(infer_server_response['generated_action']['content'])
            except Exception as e:
                print("Error parsing generated action:", e)
                import pdb; pdb.set_trace()
                
            if next_action['action'] == "STOP":
                final_sc, final_sc_path = get_screenshot(selected_screen=selected_screen)
                output_callback(
                    f"No more actions. Task complete. Final State:\n<img src='data:image/png;base64,{encode_image(str(final_sc_path))}'>",
                    sender="bot")
                break

            action_history.append(f"plan: {next_plan} - action: {next_action};")

            for message in executor({"role": "assistant", "content": next_action}, action_history):
                yield message
            
            step_count += 1
            
    else:
        raise ValueError("Invalid model selected.")


# Callback placeholders
def output_callback(response, sender):
    # print(f"[output_callback]: {response}")
    pass

def api_response_callback(response, sender):
    # print(f"[api_response_callback]: {response}")

def tool_output_callback(tool_result, sender):
    # print(f"[tool_output_callback]: {tool_result}")
    pass



if __name__ == "__main__":

    time.sleep(2)

    task = "Help me to complete the extraction of the viewer data of Downald Trump's first video on youtube,\
            fill in the excel sheet."

    api_keys = load_from_storage("api_key.json")

    print("Start task:", task)

    sampling_loop = simple_teachmode_sampling_loop(
        model="teach-mode-gpt-4o",
        task=task,
        output_callback=output_callback,
        tool_output_callback=tool_output_callback,
        api_response_callback=api_response_callback,
        selected_screen=0,
        user_id="liziqi",
        trace_id="default_trace",
        api_keys=api_keys,
    )

    for step in sampling_loop:
        print(step)
        time.sleep(1)

    print(f"Task: {task} completed. Thanks for using Teachmode-OOTB.")


import asyncio
from typing import Any, Dict, cast
from collections.abc import Callable
from anthropic.types.beta import (
    BetaContentBlock,
    BetaContentBlockParam,
    BetaImageBlockParam,
    BetaMessage,
    BetaMessageParam,
    BetaTextBlockParam,
    BetaToolResultBlockParam,
)
from anthropic.types import TextBlock
from anthropic.types.beta import BetaMessage, BetaTextBlock, BetaToolUseBlock
from ...tools import BashTool, ComputerTool, EditTool, ToolCollection, ToolResult


class AnthropicExecutor:
    def __init__(
        self, 
        output_callback: Callable[[BetaContentBlockParam], None], 
        tool_output_callback: Callable[[Any, str], None],
        selected_screen: int = 0
    ):
        self.tool_collection = ToolCollection(
            ComputerTool(selected_screen=selected_screen),
            BashTool(),
            EditTool(),
        )
        self.output_callback = output_callback
        self.tool_output_callback = tool_output_callback

    def __call__(self, response: BetaMessage, messages: list[BetaMessageParam]):
        new_message = {
            "role": "assistant",
            "content": cast(list[BetaContentBlockParam], response.content),
        }
        if new_message not in messages:
            messages.append(new_message)
        else:
            print("new_message already in messages, there are duplicates.")
        
        tool_result_content: list[BetaToolResultBlockParam] = []
        for content_block in cast(list[BetaContentBlock], response.content):
            self.output_callback(content_block)

            # Execute the tool
            if content_block.type == "tool_use":
                # Run the asynchronous tool execution in a synchronous context
                result = asyncio.run(self.tool_collection.run(
                    name=content_block.name,
                    tool_input=cast(dict[str, Any], content_block.input),
                ))
                tool_result_content.append(
                    _make_api_tool_result(result, content_block.id)
                )
                self.tool_output_callback(result, content_block.id)

            # Craft messages based on the content_block
            # Note: to display the messages in the gradio, you should organize the messages in the following way (user message, bot message)
            display_messages = _message_display_callback(messages)
            
            # Send the messages to the gradio
            for user_msg, bot_msg in display_messages:
                yield [user_msg, bot_msg], tool_result_content

        if not tool_result_content:
            return messages
        
        return tool_result_content

def _message_display_callback(messages):
    display_messages = []
    for msg in messages:
        try:
            if isinstance(msg["content"][0], TextBlock):
                display_messages.append((msg["content"][0].text, None))  # User message
            elif isinstance(msg["content"][0], BetaTextBlock):
                display_messages.append((None, msg["content"][0].text))  # Bot message
            elif isinstance(msg["content"][0], BetaToolUseBlock):
                display_messages.append((None, f"Tool Use: {msg['content'][0].name}\nInput: {msg['content'][0].input}"))  # Bot message
            elif isinstance(msg["content"][0], Dict) and msg["content"][0]["content"][-1]["type"] == "image":
                display_messages.append((None, f'<img src="data:image/png;base64,{msg["content"][0]["content"][-1]["source"]["data"]}">'))  # Bot message
            elif isinstance(msg["content"][0], Dict) and msg["content"][0]["content"][-1]["type"] == "text":
                # image_path = decode_base64_image_and_save(msg["content"][0]["content"][-1]["source"]["data"])
                # res.append((None, gr.Image(image_path)))  # Bot message
                display_messages.append((None, msg["content"][0]["content"][-1]["text"]))  # Bot message
            else:
                print(msg["content"][0])
        except Exception as e:
            print("error", e)
            pass
    return display_messages

def _make_api_tool_result(
    result: ToolResult, tool_use_id: str
) -> BetaToolResultBlockParam:
    """Convert an agent ToolResult to an API ToolResultBlockParam."""
    tool_result_content: list[BetaTextBlockParam | BetaImageBlockParam] | str = []
    is_error = False
    if result.error:
        is_error = True
        tool_result_content = _maybe_prepend_system_tool_result(result, result.error)
    else:
        if result.output:
            tool_result_content.append(
                {
                    "type": "text",
                    "text": _maybe_prepend_system_tool_result(result, result.output),
                }
            )
        if result.base64_image:
            tool_result_content.append(
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/png",
                        "data": result.base64_image,
                    },
                }
            )
    return {
        "type": "tool_result",
        "content": tool_result_content,
        "tool_use_id": tool_use_id,
        "is_error": is_error,
    }


def _maybe_prepend_system_tool_result(result: ToolResult, result_text: str):
    if result.system:
        result_text = f"<system>{result.system}</system>\n{result_text}"
    return result_text


"""
Agentic sampling loop that calls the Anthropic API and local implementation of computer use tools.
"""
import time
import json
import asyncio
import platform
from collections.abc import Callable
from datetime import datetime
from enum import StrEnum
from typing import Any, cast, Dict

from anthropic import Anthropic, AnthropicBedrock, AnthropicVertex, APIResponse
from anthropic.types import (
    ToolResultBlockParam,
    TextBlock,
)
from anthropic.types.beta import (
    BetaContentBlock,
    BetaContentBlockParam,
    BetaImageBlockParam,
    BetaMessage,
    BetaMessageParam,
    BetaTextBlockParam,
    BetaToolResultBlockParam,
)
from computer_use_ootb_internal.computer_use_demo.tools import BashTool, ComputerTool, EditTool, ToolCollection, ToolResult

import torch

from computer_use_ootb_internal.computer_use_demo.gui_agent.anthropic_agent import AnthropicActor
from computer_use_ootb_internal.computer_use_demo.executor.anthropic_executor import AnthropicExecutor
from computer_use_ootb_internal.computer_use_demo.gui_agent.planner.api_vlm_planner import APIVLMPlanner
from computer_use_ootb_internal.computer_use_demo.gui_agent.planner.local_vlm_planner import LocalVLMPlanner
from computer_use_ootb_internal.computer_use_demo.gui_agent.planner.teach_mode_vlm_planner import TeachModeVLMPlanner, split_button_reference

from computer_use_ootb_internal.computer_use_demo.gui_agent.showui_agent import ShowUIActor
from computer_use_ootb_internal.computer_use_demo.gui_agent.actor.llm_actor import LLMActor
from computer_use_ootb_internal.computer_use_demo.gui_agent.gui_parser.simple_parser.gui_parser import parse_gui

from computer_use_ootb_internal.computer_use_demo.executor.showui_executor import ShowUIExecutor
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_showui, colorful_text_vlm
from computer_use_ootb_internal.computer_use_demo.tools.screen_capture import get_screenshot
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import encode_image


BETA_FLAG = "computer-use-2024-10-22"


class APIProvider(StrEnum):
    ANTHROPIC = "anthropic"
    BEDROCK = "bedrock"
    VERTEX = "vertex"
    OPENAI = "openai"
    QWEN = "qwen"


PROVIDER_TO_DEFAULT_MODEL_NAME: dict[APIProvider, str] = {
    APIProvider.ANTHROPIC: "claude-3-5-sonnet-20241022",
    APIProvider.BEDROCK: "anthropic.claude-3-5-sonnet-20241022-v2:0",
    APIProvider.VERTEX: "claude-3-5-sonnet-v2@20241022",
    # APIProvider.OPENAI: "gpt-4o",
    # APIProvider.QWEN: "qwen2vl",
}


def sampling_loop_sync(
    *,
    model: str,
    provider: APIProvider | None,
    system_prompt_suffix: str,
    messages: list[BetaMessageParam],
    output_callback: Callable[[BetaContentBlock], None],
    tool_output_callback: Callable[[ToolResult, str], None],
    api_response_callback: Callable[[APIResponse[BetaMessage]], None],
    api_key: str,
    only_n_most_recent_images: int | None = None,
    max_tokens: int = 4096,
    selected_screen: int = 0,
    user_id: str,
    trace_id: str
):
    """
    Synchronous agentic sampling loop for the assistant/tool interaction of computer use.
    """
    
    if torch.cuda.is_available(): device = torch.device("cuda")
    elif torch.backends.mps.is_available(): device = torch.device("mps")
    else: device = torch.device("cpu") # support: 'cpu', 'mps', 'cuda'
    print(f"Model inited on device: {device}.")
    

    # TODO: More advanced way for building the model
    if model == "claude-3-5-sonnet-20241022":
        # Register Actor and Executor
        actor = AnthropicActor(
            model=model, 
            provider=provider, 
            system_prompt_suffix=system_prompt_suffix, 
            api_key=api_key, 
            api_response_callback=api_response_callback,
            max_tokens=max_tokens,
            only_n_most_recent_images=only_n_most_recent_images,
            selected_screen=selected_screen
        )

        # from IPython.core.debugger import Pdb; Pdb().set_trace()
        executor = AnthropicExecutor(
            output_callback=output_callback,
            tool_output_callback=tool_output_callback,
            selected_screen=selected_screen
        )
    elif model in ["gpt-4o + ShowUI", "qwen2-vl-max + ShowUI", "qwen-vl-7b-instruct + ShowUI"]:
        
        if model == "qwen-vl-7b-instruct + ShowUI":
            planner = LocalVLMPlanner(
                model=model,
                provider=provider,
                system_prompt_suffix=system_prompt_suffix,
                api_response_callback=api_response_callback,
                selected_screen=selected_screen,
                output_callback=output_callback,
                device=device
            )
        else:
            planner = APIVLMPlanner(
                model=model,
                provider=provider,
                system_prompt_suffix=system_prompt_suffix,
                api_key=api_key,
                api_response_callback=api_response_callback,
                selected_screen=selected_screen,
                output_callback=output_callback,
                device=device
            )
        
        actor = ShowUIActor(
            model_path="./showui-2b/",  
            # Replace with your local path, e.g., "C:\\code\\ShowUI-2B", "/Users/your_username/ShowUI-2B/".
            device=device,  
            split='web',  # 'web' or 'phone'
            selected_screen=selected_screen,
            output_callback=output_callback,
        )
        
        executor = ShowUIExecutor(
            output_callback=output_callback,
            tool_output_callback=tool_output_callback,
            selected_screen=selected_screen
        )
    
    # TODO: More modularized selection
    # Unified Model / Modularized Model - Planner -  Actor
    elif model in ["teach-mode-gpt-4o-mini", "teach-mode-qwen-vl-7b-instruct"]:
        planner = TeachModeVLMPlanner(model = "gpt-4o",  
                system_prompt_suffix="", 
                output_callback=output_callback, 
                api_response_callback=api_response_callback,
                user_id=user_id, 
                trace_id=trace_id,
                max_tokens=500,
                only_n_most_recent_images=None,
                selected_screen=0,
                print_usage=True)


        actor = LLMActor(model, None)
        
        executor = AnthropicExecutor(
            output_callback=output_callback,
            tool_output_callback=tool_output_callback,
            selected_screen=selected_screen
        )
    else:
        raise ValueError(f"Model {model} not supported")
    
    print(f"Model Inited: {model}, Provider: {provider}")
    
    tool_result_content = None
    
    print(f"Start the message loop. User messages: {messages}")
    
    if model == "claude-3-5-sonnet-20241022": # Anthropic loop
        while True:
            response = actor(messages=messages)

            for message, tool_result_content in executor(response, messages):
                yield message
        
            if not tool_result_content:
                return messages

            messages.append({"content": tool_result_content, "role": "user"})
    
    elif model == "gpt-4o + ShowUI" or model == "qwen2-vl-max + ShowUI" or model == "qwen-vl-7b-instruct + ShowUI":  # ShowUI loop 
        while True:
            vlm_response = planner(messages=messages)
            
            next_action = json.loads(vlm_response).get("Next Action")
            yield next_action
            
            if next_action == None or next_action == "" or next_action == "None":
                final_sc, final_sc_path = get_screenshot(selected_screen=selected_screen)
                output_callback(f'No more actions from {colorful_text_vlm}. End of task. Final State:\n<img src="data:image/png;base64,{encode_image(str(final_sc_path))}">',
                                sender="bot")
                yield None
                        
            output_callback(f"{colorful_text_vlm} sending action to {colorful_text_showui}:\n{next_action}", sender="bot")
            
            actor_response = actor(messages=next_action)
            yield actor_response
            
            for message, tool_result_content in executor(actor_response, messages):
                time.sleep(0.5)
                yield message
                
            # since showui executor has no feedback for now, we use "actor_response" to represent its response
            # update messages for the next loop
            messages.append({"role": "user",
                             "content": ["History plan:" + str(json.loads(vlm_response)) + 
                                         "History actions:" + str(actor_response["content"])]
                             })
            print(f"End of loop. Messages: {str(messages)[:100000]}. Total cost: $USD{planner.total_cost:.5f}")

    elif model in ["teach-mode-gpt-4o-mini", "teach-mode-qwen-vl-7b-instruct"]: # Teach mode loop
        while True:
            # TODO: Only support Windows Platform for now, check the platform
            if platform.system() != "Windows":
                raise ValueError("Teach mode only supports Windows Platform for now.")

            vlm_response = planner(messages=messages)

            next_action = json.loads(vlm_response)['Next Action']
            next_action = split_button_reference(next_action)

            yield next_action

            # Screenshot for final step
            if next_action == None or next_action == "" or next_action == "None":
                final_sc, final_sc_path = get_screenshot(selected_screen=selected_screen)
                output_callback(f'No more actions from {colorful_text_vlm}. End of task. Final State:\n<img src="data:image/png;base64,{encode_image(str(final_sc_path))}">',
                                sender="bot")
                yield None

            output_callback(f"{colorful_text_vlm} sending action to ootb-small:\n{next_action}", sender="bot")

            # TODO: optimize the screenshot capture, too many times
            uia_data, screenshot_path = get_screenshot(selected_screen=selected_screen)

            # TODO: only gui as output
            gui = parse_gui(user_id=user_id, trace_id=trace_id, screenshot_path=screenshot_path, user_scaleFactor="auto", uia_data=uia_data, query=next_action, mode="teach", ocr_mode="googleocr")
            actor_response = actor(next_action, gui, screenshot_path)
            
            yield actor_response

            for message, tool_result_content in executor(actor_response, messages):
                time.sleep(0.5)
                yield message

            messages.append({"role": "user",
                             "content": ["History plan:" + str(json.loads(vlm_response)) + 
                                        "History actions:" + str(actor_response["content"])]
                            })
            print(f"End of loop. Messages: {str(messages)[:100000]}. Total cost: $USD{planner.total_cost:.5f}")



import cv2
import numpy as np
import datetime
import importlib
import textdistance
from PIL import Image

# from autopc.base_module import BaseModule
from .text_detection import text_detection
from .icon_detection.icon_detection import detect_icons
from .utils import *

# from ultralytics import YOLO
# from config import autopc_config

# from ui_text_detection import text_detection
# from button_detection import detect_button
# from utils import crop_panel, restore_coordinate, get_current_time, is_in_bbox


class GUIParserBase:
    name = "gui_parser"
    description = """
This tool can extract the information of screenshot.
Invoke command: gui_parser(query, visual[i])
:param query -> str, specific command. visual[i] -> image, the latest screenshot.
"""

    def __init__(self):
        # judge if the cache folder exists
        super(GUIParserBase, self).__init__()
        # self.cache_folder = cache_folder
        # self.task_id = get_current_time()
        # YOLOv8 model that can detect 600 classes
        # self.yolo_model = self.build_model("yolov8")
        # self.parsers = []

    def _run(self, meta_data, screenshot_path, software_name=None):
        # get main panel
        self.software_name = software_name
        self.screenshot_paths.append(screenshot_path)
        self.parsed_gui = {"panel": []}
        if self.software_name in [
            "Calculator",
            "计算器",
            "file explorer",
            "任务栏",
            "Taskbar",
            "Xshell",
            "RStudio",
            "Zoom",
            "Tecent meeting",
            "Power BI",
            "web",
        ]:
            self.parsed_gui = self.get_panel_uia(meta_data)
            # Template matching for these apllications
            for panel_item in self.parsed_gui["panel"]:
                if panel_item["name"] not in ["Title Bar", "Navigation Bar"]:
                    button_box = self.get_button(panel_item, self.screenshot_paths[-1])
                    panel_item["elements"] += button_box
            return self.parsed_gui

        print("Screenshot:", screenshot_path)
        highlight_ocr = self.detect_highlight_with_ocr(screenshot_path)
        _, ocr = text_detection(screenshot_path, save_png=False)
        for window_name, window_meta_data in meta_data.items():
            if not window_meta_data:
                continue
            if window_meta_data[0]["properties"]["friendly_class_name"] in [
                "MenuItem",
                "Edit",
                "Button",
            ]:
                self.parsed_gui["panel"] += self.get_popup_window(
                    window_meta_data, window_name
                )
            else:
                self.parsed_gui["panel"] += self.get_panel(
                    window_meta_data, ocr, highlight_ocr
                )
                self.parsed_gui["panel"] += self.get_menu(window_meta_data)
                # TODO: get the title bar
        return self.parsed_gui

    def reset(self):
        self.screenshot_paths = []
        self.task_id = get_current_time()

    def postprocess_uia(self, metadata):
        # iterate each window
        for window_data in metadata.values():

            for element_data in window_data:
                # get all elements of current level
                elements = element_data.get("elements", [])

                # if the class name is TitleBar, regain the rectangle of the whole panel
                if element_data["class_name"] == "TitleBar" and len(elements) > 0:
                    left = min(e.get("rectangle", [0, 0, 0, 0])[0] for e in elements)
                    top = min(e.get("rectangle", [0, 0, 0, 0])[1] for e in elements)
                    right = max(e.get("rectangle", [0, 0, 0, 0])[2] for e in elements)
                    bottom = max(e.get("rectangle", [0, 0, 0, 0])[3] for e in elements)
                    element_data["rectangle"] = [left, top, right, bottom]
                # sort elements by x and y
                element_data["elements"] = sort_elements_by_xy(elements)

        return metadata

    # retreieve the panel meta data directively from UIA (pywinauto)

    def postprocess_icon(self, icon_items, screenshot_path):
        image = Image.open(screenshot_path)
        width, height = image.size
        result = {
            "GUI": [
                {
                    "name": "Main Content",
                    "rectangle": [0, 0, width, height],
                    "elements": icon_items,
                }
            ]
        }
        return result

    def get_panel_uia(self, control_info_list, screenshot_path):
        text2panel_name = {
            "新建会话属性": "New session properties",
            "会话管理器": "Session manager",
        }

        def recurse_controls(control_info, dialog_components, depth):
            children = control_info["children"]
            if len(children) == 0:
                return
            for child_control in children:
                child_properties = child_control["properties"]
                child_friendly_class_name = child_properties["friendly_class_name"]
                child_properties_name = ""
                # Get the feasible name of the child control
                if len(child_properties["texts"]) == 0:
                    child_properties_name = ""
                elif (
                    isinstance(child_properties["texts"][0], list)
                    and len(child_properties["texts"]) != 0
                ):
                    result = []
                    for item in child_properties["texts"]:
                        if isinstance(item[0], str):
                            result.append("".join(item))
                    child_properties_name = "".join(result)
                else:
                    child_properties_name = child_properties["texts"][0]
                # Get the search bar
                # or 'search' not in child_properties_name.lower()
                if child_friendly_class_name in ["Edit", "ComboBox"] and (
                    child_properties_name == ""
                ):
                    child_properties_name = "Search Bar"
                # Rules to exclude some class name
                # if self.software_name == 'Xshell':
                #     exclude_class_name_list = ['GroupBox', 'ListBox', 'Custom', 'Menu', 'Pane', 'Toolbar', 'TabControl', 'DataItem']
                # elif self.software_name == 'RStudio':
                #     exclude_class_name_list = ['GroupBox', 'ListBox', 'Custom', 'Menu', 'Pane', 'Toolbar', 'TabControl', 'DataItem', 'TabItem', 'MenuItem', 'Button', 'Document']
                # elif self.software_name == 'web video':
                #     exclude_class_name_list = ['GroupBox', 'ListBox', 'ListItem', 'Custom', 'Menu', 'Pane', 'Toolbar', 'TabControl', 'TreeItem', 'DataItem']
                # else:
                #     exclude_class_name_list = ['GroupBox', 'ListBox', 'ListItem', 'Custom', 'Menu', 'Pane', 'Toolbar', 'TabControl', 'TreeItem', 'DataItem', 'Hyperlink']
                if (
                    (
                        "search" in child_properties_name.lower()
                        and child_friendly_class_name == "GroupBox"
                    )
                    or (
                        child_friendly_class_name not in self.exclude_class_name_list
                        or len(child_control["children"]) == 0
                    )
                    and (child_properties_name not in ["", '"'])
                    and (
                        not all(
                            element == 0 for element in child_properties["rectangle"]
                        )
                    )
                ):

                    left, top, right, bottom = child_properties["rectangle"]
                    left_bound, top_bound, right_bound, bottom_bound = (
                        dialog_components["rectangle"]
                    )

                    if self.software_name not in [
                        "word",
                        "excel",
                        "powerpoint",
                        "file explorer",
                    ]:
                        if left < left_bound:
                            child_properties["rectangle"][0] = left_bound
                        if top < top_bound:
                            child_properties["rectangle"][1] = top_bound
                        if right > right_bound:
                            child_properties["rectangle"][2] = right_bound
                        if bottom > bottom_bound:
                            child_properties["rectangle"][3] = bottom_bound
                    if not (
                        child_properties["rectangle"][0]
                        >= child_properties["rectangle"][2]
                        or child_properties["rectangle"][1]
                        >= child_properties["rectangle"][3]
                    ):
                        # TODO: This involves an operation of cropping the panel image, which is time-consuming.
                        # panel_img = crop_panel(child_properties['rectangle'], screenshot_path)
                        # TODO: delete it| or not np.all(panel_img == panel_img[0, 0]):
                        if (child_friendly_class_name in ["CheckBox"] or True) and len(
                            child_properties_name
                        ) <= 200:
                            # remove unseen characters
                            child_properties_name = child_properties_name.replace(
                                "\u200b", ""
                            )
                            dialog_components["elements"].append(
                                {
                                    "name": child_properties_name,
                                    "rectangle": child_properties["rectangle"],
                                    "class_name": child_friendly_class_name,
                                    "type": ["Click", "rightClick"],
                                    "depth": depth + "-" + str(self.count),
                                }
                            )
                            self.count += 1
                # Post-processing for Outlook
                if (
                    child_friendly_class_name in ["Button", "MenuItem", "TabItem"]
                    and self.software_name == "outlook"
                ):
                    continue

                recurse_controls(child_control, dialog_components, depth)

        main_name = list(control_info_list.keys())[0]
        dialog_components = {self.software_name: []}
        excel_iter = iter(
            ["Navigation Bar", "Tools Bar", "Function Bar", "Main Content"]
        )
        # if any(keyword in main_name for keyword in ['哔哩哔哩', 'bilibili', '腾讯视频']):
        #     self.software_name = 'web video'
        # Check if the friendly_class_name
        for control_info in control_info_list[main_name]:
            if (
                control_info["properties"]["friendly_class_name"]
                in [
                    "Dialog",
                    "Pane",
                    "GroupBox",
                    "TitleBar",
                    "Menu",
                    "Document",
                    "ListBox",
                ]
                and len(control_info["children"]) != 0
            ):
                # Append texts and rectangle to the Dialog components
                # if all(element <= 0 for element in control_info['properties']['rectangle']) and \
                #             control_info['properties']['friendly_class_name'] not in ['TitleBar']:
                #     continue
                # Main panel name rules
                control_name = control_info["properties"]["texts"][0]
                if control_info["properties"]["texts"][0] == "":
                    if control_info["properties"]["friendly_class_name"] == "TitleBar":
                        control_name = "Title Bar"
                    if control_info["properties"]["friendly_class_name"] == "Document":
                        control_name = "Main Content"

                    if control_info["properties"]["friendly_class_name"] == "Pane":
                        if self.software_name in ["web", "web video"]:
                            control_name = "Navigation Bar"
                        else:
                            if self.software_name == "excel":
                                control_name = next(excel_iter, "Main Content")
                            else:
                                control_name = "Main Content"
                else:
                    if control_info["properties"][
                        "friendly_class_name"
                    ] == "Document" and self.software_name in ["web", "web video"]:
                        control_name = "Main Content"
                        # if 'Outlook' in main_name and 'Mail' in main_name:
                        #     original_value = dialog_components.pop(self.software_name)
                        #     self.software_name = 'Outlook'
                        #     dialog_components['Outlook'] = original_value
                    else:
                        control_name = control_info["properties"]["texts"][0]
                if control_name in text2panel_name.keys():
                    control_name = text2panel_name[control_name]
                self.count = 1
                dialog_components[self.software_name].append(
                    {
                        "name": control_name,
                        "rectangle": control_info["properties"]["rectangle"],
                        "class_name": control_info["properties"]["friendly_class_name"],
                        "depth": "1",
                        "elements": [],
                    }
                )
                # Process children of the Dialog
                recurse_controls(
                    control_info, dialog_components[self.software_name][-1], "1"
                )
        return dialog_components

    @staticmethod
    def merge_elements(panel_item):
        # print(panel_item)
        empty = True
        for k, v in panel_item.items():
            if v:
                empty = False
                break
        if empty:
            return []
        merged_control, elements_to_be_merged = [], []
        for key, value in panel_item.items():
            if not merged_control:
                # if key == 'button':
                #     merged_control = [row for row in panel_item[key]]
                # else:
                #     merged_control = [list(row) for row in panel_item[key]]
                if key == "editing_control":
                    merged_control = [list(row) for row in panel_item[key]]
                else:
                    elements_to_be_merged += panel_item.get(key, [])
            else:
                elements_to_be_merged += panel_item.get(key, [])

        for button in elements_to_be_merged:
            y_center = (button["rectangle"][1] + button["rectangle"][3]) / 2
            row_index = find_appropriate_row(merged_control, y_center)
            if row_index is not None:
                insert_into_row(merged_control[row_index], button)
            else:
                # If button doesn't fit into any row, append it as a new row.
                merged_control.append([button])
        # Sort rows from top to bottom based on the center y-coordinate of the first element
        try:
            merged_control.sort(
                key=lambda row: (
                    (row[0]["rectangle"][1] + row[0]["rectangle"][3]) / 2,
                    (row[0]["rectangle"][0] + row[0]["rectangle"][2]) / 2,
                )
            )
        except:
            print(merged_control)
        return merged_control

    @staticmethod
    def get_menu(meta_data):
        # get menu item
        menu_item = []
        for item in meta_data:
            if item["properties"]["friendly_class_name"] == "Menu":
                # only keep one level children
                new_item = {
                    "name": item["properties"]["texts"][0],
                    "rectangle": item["properties"]["rectangle"],
                    "elements": [
                        [
                            {
                                "name": child["properties"]["texts"][0],
                                "rectangle": child["properties"]["rectangle"],
                            }
                            for child in item["children"]
                            if child["properties"]["texts"]
                        ]
                    ],
                }
                menu_item.append(new_item)
        return menu_item


    
    def get_button(self, panel_item, screenshot_path, scale_factor="1.0x"):
        # print("GXW",scale_factor)
        # crop the panel based on the rectangle
        # print("processing button fuck")
        print(panel_item["name"], panel_item["rectangle"])
        panel_img = crop_panel(panel_item["rectangle"], screenshot_path)
        # get the button
        # th = 0.90 if self.software_name == 'after effect' else 0.78

        th = 0.80 if self.software_name == "after effect" else 0.75
        if self.software_name in [
            "premiere",
            "after effect",
            "Premiere",
            "After Effect",
        ]:
            print("Screenshot:", screenshot_path)
            button_box = detect_button_pr_ae(
                panel_img,
                software_name=self.software_name,
                panel_name=panel_item["name"],
                threshold=th,
                scale_factor=scale_factor,
            )
        else:
            button_box = detect_button(
                panel_img,
                software_name=self.software_name,
                panel_name=panel_item["name"],
                threshold=th,
            )

        if self.software_name == "after effect" and panel_item["name"] == "Timeline":
            hsv = cv2.cvtColor(panel_img, cv2.COLOR_BGR2HSV)
            # 定义灰色的 HSV 范围
            lower_gray = np.array([0, 0, 46])
            upper_gray = np.array([180, 40, 220])
            # 创建掩膜，将在指定颜色范围内的部分设置为白色，其他部分设置为黑色
            mask = cv2.inRange(hsv, lower_gray, upper_gray)
            # 将原始图像与掩膜进行按位与操作，提取对应部分
            result = cv2.bitwise_and(panel_img, panel_img, mask=mask)
            result = process_image_highlight_gray(result)
            result = np.repeat(result[..., np.newaxis], 3, 2)
            save_path = self.cache_folder + "gray.png"
            cv2.imwrite(save_path, result)
            check_box = detect_button(
                result,
                software_name=self.software_name,
                panel_name=panel_item["name"],
                threshold=0.9,
            )
            button_box += check_box
            print(check_box, panel_item["name"])
        # restore the button coordinate to the whole screenshot
        button_box = restore_coordinate(button_box, panel_item["rectangle"])
        return button_box

    @staticmethod
    def get_text(panel_item, ocr, screenshot_path):
        # Step1: Find all the texts in the panel
        panel_texts = [item for item in ocr["texts"]]
        if not panel_texts:
            return []
        # obtain information about panel texts
        sorted_panel_texts = sorted(
            panel_texts, key=lambda x: (x["bbox"][1], x["bbox"][0])
        )
        # Step 2: Identify rows by grouping elements that are in approximately the same vertical position.
        editing_controls = []
        current_row = []
        previous_y = sorted_panel_texts[0]["bbox"][1]

        count = 1
        for item in sorted_panel_texts:
            y1 = item["bbox"][1]
            # If the vertical position of the current item is significantly different from the previous one, start a new row.
            # You might need to adjust this threshold based on your panel_texts.
            if abs(y1 - previous_y) > 15:
                editing_controls.append(current_row)
                current_row = []

            current_row.append(
                {
                    "name": item["content"],
                    "rectangle": item["bbox"],
                }
            )
            previous_y = y1
        # Add the last row if not empty.
        if current_row:
            editing_controls.append(current_row)
        # Step 3: Sort elements within each row based on their horizontal position.
        for i, row in enumerate(editing_controls):
            editing_controls[i] = sorted(row, key=lambda x: x["rectangle"][0])
        return editing_controls

    def get_editing_control_icons(self, panel_item, screenshot_path):
        # crop the panel based on the rectangle
        panel_img = crop_panel(panel_item["rectangle"], screenshot_path)
        # get the button
        button_box = detect_button(
            panel_img,
            software_name=self.software_name,
            panel_name=panel_item["name"],
            icon_type="editing_control",
        )
        # restore the button coordinate to the whole screenshot
        button_box = restore_coordinate(button_box, panel_item["rectangle"])
        return button_box

    @staticmethod
    def get_scroll(panel_item, button_list, ocr, screenshot_path):
        if panel_item["name"] not in ["Lumetri Color"]:
            return []
        # TODO: first detect the circle, then use 霍夫变化 to detect the line?
        scroll_button = [b for b in button_list if "scroll bar" in b["name"]]
        scroll_text = [
            t for t in ocr["texts"] if is_in_bbox(t["bbox"], panel_item["rectangle"])
        ]
        scroll_bar_items = []
        for button in scroll_button:
            l, t, r, b = button["rectangle"]
            cur_bar = {"Button center": ((l + r) // 2, (t + b) // 2)}
            rectangle = [None, None, None, None]
            for text in scroll_text:
                # Find scroll bar name
                if (
                    0 < abs(text["bbox"][1] - button["rectangle"][1]) < 10
                    and 0 < button["rectangle"][0] - text["bbox"][0] < 500
                ):
                    cur_bar["name"] = text["content"]
                    rectangle[0] = text["bbox"][0]
                    rectangle[1] = text["bbox"][1]
                    rectangle[3] = text["bbox"][3]
                # Find current value of the scroll bar
                elif (
                    0 < abs(text["bbox"][1] - button["rectangle"][1]) < 10
                    and 0 < text["bbox"][0] - button["rectangle"][0] < 500
                ):
                    cur_bar["cur_value"] = text["content"]
                    rectangle[2] = text["bbox"][2]
            # Detect lines use houghLinesP
            scroll_image = crop_panel(rectangle, screenshot_path)
            gray = cv2.cvtColor(scroll_image, cv2.COLOR_BGR2GRAY)
            edges = cv2.Canny(gray, 50, 150, apertureSize=3)
            lines = cv2.HoughLinesP(
                edges, 1, np.pi / 180, 100, minLineLength=100, maxLineGap=10
            )
            if lines is not None and len(lines) > 0:
                for line in lines:
                    x1, y1, x2, y2 = line[0]
                    scroll_bar = [x1, y1, x2, y2]
                    if None in rectangle:
                        break
                    else:
                        cur_bar["rectangle"] = scroll_bar
                        cur_bar["type"] = ["click"]
                        cur_bar = restore_coordinate([cur_bar], rectangle)[0]
                        scroll_bar_items.append(cur_bar)
                    # cv2.line(scroll_image, (x1, y1), (x2, y2), (0, 255, 0), 2)
                    # cv2.imwrite("/home/aiassist/difei/assistgui/.cache/scroll.png", scroll_image)
                    break
            # scroll_bar_items.append(cur_bar)
        return scroll_bar_items

    def get_drag_position_pr(self, panel_img, cache_folder):
        hsv = cv2.cvtColor(panel_img, cv2.COLOR_BGR2HSV)
        # Define the scope of gray
        lower_gray = np.array([0, 0, 40])
        # upper_gray = np.array([0, 0, 50])
        upper_gray = np.array([0, 5, 55])
        mask = cv2.inRange(hsv, lower_gray, upper_gray)
        result = cv2.bitwise_and(panel_img, panel_img, mask=mask)
        non_blue_part = cv2.bitwise_not(mask)
        result[non_blue_part > 0] = 255
        cache_folder = autopc_config["os_agent_settings"]["cache_dir"]
        save_path = os.path.join(cache_folder, "cur_pane_gray.png")
        cv2.imwrite(save_path, result)
        panel_img = cv2.imread(save_path)
        gray = cv2.cvtColor(panel_img, cv2.COLOR_BGR2GRAY)
        edges = cv2.Canny(gray, 50, 150, apertureSize=3)
        lines = cv2.HoughLinesP(
            edges, 1, np.pi / 180, 100, minLineLength=100, maxLineGap=10
        )

        if lines is not None and len(lines) > 0:
            lines = np.squeeze(lines, axis=1)
            # find the first and last horizontal line
            horizontal_lines = lines[
                (lines[:, 1] == lines[:, 3]) & (lines[:, 0] < lines[:, 2])
            ]
            if len(horizontal_lines) != 0:
                horizontal_lines = horizontal_lines[np.argsort(horizontal_lines[:, 1])]
                first_horizontal_line = horizontal_lines[0]
                last_horizontal_line = horizontal_lines[-1]
            # find the first and last vertical line
            vertical_lines = lines[
                (lines[:, 0] == lines[:, 2]) & (lines[:, 3] < lines[:, 1])
            ]
            if len(vertical_lines) != 0:
                vertical_lines = vertical_lines[np.argsort(vertical_lines[:, 0])]
                first_vertical_line = vertical_lines[0]
                last_vertical_line = vertical_lines[-1]
            # return the final bounding box
            bounding_box = None
            if len(horizontal_lines) != 0 and len(vertical_lines) != 0:
                # bounding_box = [first_vertical_line[0] + panel_item['rectangle'][0], first_horizontal_line[1] + panel_item['rectangle'][1],
                #                 last_vertical_line[2] + panel_item['rectangle'][2], last_horizontal_line[3] + panel_item['rectangle'][3]]
                bounding_box = [
                    first_vertical_line[0],
                    first_horizontal_line[1],
                    last_vertical_line[2],
                    last_horizontal_line[3],
                ]
                cv2.rectangle(
                    panel_img,
                    (first_vertical_line[0], first_horizontal_line[1]),
                    (last_vertical_line[2], last_horizontal_line[3]),
                    (0, 255, 0),
                    2,
                )
                cv2.imwrite(
                    os.path.join(cache_folder, "cur_pane_gray_line.png"), panel_img
                )

            return bounding_box

    def get_drag_position_ae(self, panel_img):
        hsv = cv2.cvtColor(panel_img, cv2.COLOR_BGR2HSV)
        # Define the scope of gray
        lower_gray = np.array([0, 0, 0])
        upper_gray = np.array([20, 20, 20])
        mask = cv2.inRange(hsv, lower_gray, upper_gray)
        result = cv2.bitwise_and(panel_img, panel_img, mask=mask)
        non_blue_part = cv2.bitwise_not(mask)
        result[non_blue_part > 0] = 255
        cache_folder = autopc_config["os_agent_settings"]["cache_dir"]
        save_path = os.path.join(cache_folder, "cur_pane_gray.png")
        cv2.imwrite(save_path, result)
        panel_img = cv2.imread(save_path)
        gray = cv2.cvtColor(panel_img, cv2.COLOR_BGR2GRAY)
        edges = cv2.Canny(gray, 50, 150, apertureSize=3)
        lines = cv2.HoughLinesP(
            edges, 1, np.pi / 180, 100, minLineLength=100, maxLineGap=10
        )

        if lines is not None and len(lines) > 0:
            lines = np.squeeze(lines, axis=1)
            # find the first and last horizontal line
            horizontal_lines = lines[
                (lines[:, 1] == lines[:, 3]) & (lines[:, 0] < lines[:, 2])
            ]
            if len(horizontal_lines) != 0:
                horizontal_lines = horizontal_lines[np.argsort(horizontal_lines[:, 1])]
                first_horizontal_line = horizontal_lines[0]
                last_horizontal_line = horizontal_lines[-1]
            # find the first and last vertical line
            vertical_lines = lines[
                (lines[:, 0] == lines[:, 2]) & (lines[:, 3] < lines[:, 1])
            ]
            if len(vertical_lines) != 0:
                vertical_lines = vertical_lines[np.argsort(vertical_lines[:, 0])]
                first_vertical_line = vertical_lines[0]
                # last_vertical_line = vertical_lines[-1]
            # return the final bounding box
            bounding_box = None
            if len(horizontal_lines) != 0 and len(vertical_lines) != 0:
                # bounding_box = [first_vertical_line[0] + panel_item['rectangle'][0], first_horizontal_line[1] + panel_item['rectangle'][1],
                #                 last_vertical_line[2] + panel_item['rectangle'][2], last_horizontal_line[3] + panel_item['rectangle'][3]]
                bounding_box = [
                    first_vertical_line[0],
                    first_horizontal_line[1],
                    first_horizontal_line[2],
                    last_horizontal_line[3],
                ]
                cv2.rectangle(
                    panel_img,
                    (first_vertical_line[0], first_horizontal_line[1]),
                    (first_horizontal_line[2], last_horizontal_line[3]),
                    (0, 255, 0),
                    2,
                )
                cv2.imwrite(
                    os.path.join(cache_folder, "cur_pane_gray_line.png"), panel_img
                )

            return bounding_box

    def get_timeline(self, panel_item, screenshot_path):
        # TODO: Locate the text display the text
        # if panel_item['name'] not in ['Timeline', 'Program', 'Accessory']:
        #     return [], None, []
        scroll_image = crop_panel(panel_item["rectangle"], screenshot_path)
        # Transform image format from BGR to HSV
        hsv = cv2.cvtColor(scroll_image, cv2.COLOR_BGR2HSV)
        # Define the scope of blue
        lower_blue = np.array([105, 100, 100])
        upper_blue = np.array([130, 255, 255])
        # Creates a mask that sets parts within a specified color range to white and other parts to black
        mask = cv2.inRange(hsv, lower_blue, upper_blue)
        # Perform a bitwise AND operation on the original image and the mask to extract the blue part
        result = cv2.bitwise_and(scroll_image, scroll_image, mask=mask)
        save_path = self.cache_folder + "blue.png"
        cv2.imwrite(save_path, result)
        _, ocr = text_detection(save_path, save_png=False)

        cv2.imwrite(self.cache_folder + "cur_pane_timeline.png", scroll_image)
        _, ocr_cur = text_detection(
            self.cache_folder + "cur_pane_timeline.png", save_png=False
        )
        gray = cv2.cvtColor(result, cv2.COLOR_BGR2GRAY)
        edges = cv2.Canny(gray, 50, 150, apertureSize=3)
        lines = cv2.HoughLinesP(
            edges, 1, np.pi / 180, 100, minLineLength=100, maxLineGap=10
        )
        timeline = []
        bbox_list = []
        playhead = None
        horizontal_line = None
        drag_position = None
        timeline_flag = False
        if lines is not None and len(lines) >= 2:
            # store the position of action dragTo
            if self.software_name == "premiere":
                drag_position = self.get_drag_position_pr(
                    scroll_image, self.cache_folder
                )
            else:
                drag_position = self.get_drag_position_ae(scroll_image)

            if drag_position is not None:
                drag_position = [
                    {"name": "dragTo_timeline", "rectangle": drag_position}
                ]
                # drag_position = restore_coordinate(drag_position, panel_item['rectangle'])
            for line in lines:
                x1, y1, x2, y2 = line[0]
                rectangle = [x1, y1, x2, y2]
                # Do not consider boarder lines
                if (
                    x1 > 5
                    and x1
                    <= x2
                    < panel_item["rectangle"][2] - panel_item["rectangle"][0] - 5
                ):
                    if playhead is None and x2 - x1 < 3:
                        timeline.append(
                            {
                                "name": "playhead",
                                "rectangle": rectangle,
                                "type": self.action_type + ["dragTo_timeline"],
                            }
                        )
                    else:
                        bbox_list.append(rectangle)

                cv2.line(result, (x1, y1), (x2, y2), (0, 255, 0), 2)
            cv2.imwrite(".cache/" + "cur_pane_line.png", result)
            if len(bbox_list) != 0:
                horizontal_line = min(bbox_list, key=lambda sublist: sublist[1])
            if playhead is not None and horizontal_line is not None:
                horizontal_line[0] = min(horizontal_line[0], playhead[0])
        # Find all the texts in the panel, filter the timelike text and in blue
        i = 1
        drag_block = []
        remove_chars = (
            "[0-9’!\"#$%&'()*+,-./:;<=>?@，。?★、…【】《》？“”‘’！[\\]^_`{|}~1.'`-]+"
        )
        for text in ocr["texts"]:
            if match_time_format(text["content"]):
                timeline.append(
                    {
                        "name": "timeline-" + str(i),
                        "rectangle": text["bbox"],
                        "cur_value": text["content"],
                        "type": self.action_type + ["typewrite", "rightClick"],
                    }
                )
                timeline_flag = True
                i += 1
            if playhead is not None and horizontal_line is not None:
                if abs(horizontal_line[1] - text["bbox"][1]) < 10:
                    timeline.append(
                        {
                            "name": text["content"] + "|layer",
                            "rectangle": horizontal_line,
                            "type": self.action_type
                            + ["doubleClick", "rightClick", "dragTo_timeline"],
                        }
                    )

        for text in ocr_cur["texts"]:
            if drag_position is not None:
                if is_in_bbox(text["bbox"], drag_position[0]["rectangle"]) and re.sub(
                    remove_chars, "", text["content"].lower()
                ) not in ["fx", "l", "r", ""]:
                    drag_block.append(
                        {"name": text["content"], "rectangle": text["bbox"]}
                    )

        timeline = restore_coordinate(timeline, panel_item["rectangle"])
        drag_block = restore_coordinate(drag_block, panel_item["rectangle"])
        if drag_position is not None:
            drag_position = restore_coordinate(drag_position, panel_item["rectangle"])

        return timeline, drag_position, drag_block, timeline_flag

    @staticmethod
    def get_search_bar(panel_item, raw_item):
        # TODO: from meta_data
        all_search_bar = []

        def search_edit_control(item):
            if item["properties"]["friendly_class_name"] == "Edit":
                all_search_bar.append(
                    {
                        "name": "Search bar",
                        "rectangle": item["properties"]["rectangle"],
                        "type": ["click", "typewrite"],
                    }
                )
            if len(item["children"]) > 0:
                for i in range(len(item["children"])):
                    search_edit_control(item["children"][i])
            return all_search_bar

        return search_edit_control(raw_item)

    @staticmethod
    def get_asset_bar(panel_item):
        if panel_item["name"] not in ["Timeline"]:
            return []
        return []

    # @staticmethod

    def get_media_asset(self, panel_item, raw_item, screenshot_path):
        if panel_item["name"] not in ["Program", "Composition", "Layer"]:
            return []
        try:
            media_asset_box = raw_item["children"][0]["children"][0]["children"][0][
                "properties"
            ]["rectangle"]
        except:
            media_asset_box = panel_item["rectangle"]

        media_asset = crop_panel(media_asset_box, screenshot_path)
        object_list = []  # self.get_objects(media_asset, media_asset_box)
        # TODO: add guideline

        return [
            {
                "name": "media_asset",
                # TODO: check if it is general for other software, only test on PR
                "rectangle": media_asset_box,
                "type": self.action_type + ["rightClick"],
                "objects": object_list,
            }
        ]
        # 'objects': object_list}]

    def get_objects(self, media_asset, media_asset_box):
        results = self.yolo_model.predict(
            source=media_asset, save=False, save_txt=False
        )
        object_list = []
        for result in results:
            boxes = result.boxes
            names = result.names
            for cls, box in zip(
                boxes.cls.cpu().numpy().astype(np.uint32),
                boxes.xyxy.cpu().numpy().astype(np.uint32),
            ):
                cur_children = {
                    "name": names[cls],
                    "rectangle": [int(x) for x in box.tolist()],
                }
                object_list.append(cur_children)
        return restore_coordinate(object_list, media_asset_box)

    @staticmethod
    def get_popup_window(meta_data, window_name):
        panel_item = {"name": window_name, "rectangle": None}
        # Filter out dictionaries where 'texts' field is empty
        filtered_data = [
            entry for entry in meta_data if entry["properties"]["texts"][0]
        ]
        # Sort the data based on the y-coordinate of the 'rectangle' field (from top to bottom)
        sorted_data_y = sorted(
            filtered_data, key=lambda x: x["properties"]["rectangle"][1]
        )
        elements = []
        current_row = []
        prev_y = None
        for entry in sorted_data_y:
            y = entry["properties"]["rectangle"][1]
            # Check if this entry should belong to a new row
            if prev_y is not None and abs(y - prev_y) > 10:
                # Sort the current row from left to right based on the x-coordinate of the 'rectangle' field
                sorted_row_x = sorted(current_row, key=lambda x: x["rectangle"][0])
                elements.append(sorted_row_x)
                current_row = []
            if entry["properties"]["friendly_class_name"] in ["MenuItem", "Button"]:
                name = f"{entry['properties']['texts'][0]}"
            else:
                name = f"{entry['properties']['texts'][0]}|{entry['properties']['friendly_class_name']}"
            current_row.append(
                {"name": name, "rectangle": entry["properties"]["rectangle"]}
            )
            prev_y = y
        if current_row:
            # Sort the last row from left to right before appending it to the matrix
            sorted_row_x = sorted(current_row, key=lambda x: x["rectangle"][0])
            elements.append(sorted_row_x)
        panel_item["elements"] = elements
        panel_item["rectangle"] = find_compact_bounding_box(elements)
        return [panel_item]

    def recognize_panel(self, item, ocr, screenshot_path):
        # TODO: peiran, should be refined with more elegant way
        # first get panel name with ocr
        # Ipython set trace
        # from IPython.core.debugger import Pdb
        # Pdb().set_trace()
        if self.software_name in ["word", "powerpoint", "excel"]:
            return item["properties"]["texts"][0]
        if self.software_name in ["adobe acrobat"]:
            if item["properties"]["texts"][0] != "AVTabsContainerView":
                self.accessory_number += 1
                return "Accessory" + str(self.accessory_number)
        success, panel_name = self.recognize_panel_with_ocr(
            item["properties"]["rectangle"], ocr["texts"]
        )
        if not success:
            success, panel_name = self.recognize_panel_with_icon(
                item["properties"]["rectangle"], screenshot_path
            )
        if not success:
            self.accessory_number += 1
            panel_name = "Accessory" + str(self.accessory_number)
        return panel_name

    def detect_highlight_with_ocr(self, screenshot_path):
        # panel_img = crop_panel(panel_bbox, screenshot_path)
        panel_img = cv2.imread(screenshot_path)
        panel_img = process_image_highlight(panel_img)
        print(f"Cache_folder: {self.cache_folder}")
        save_path = self.cache_folder + "cur_pane.png"
        cv2.imwrite(save_path, panel_img)
        _, ocr = text_detection(save_path, save_png=False)
        return ocr

    def recognize_panel_with_ocr(self, panel_bbox, ocr_result):
        # TODO: peiran, highlight the panel bbox
        """
        panel_bbox = [x1, y1, x2, y2]
        ocr_results =
        {'img_shape': (1440, 2560, 3),
         'texts': [{'content': 'Pr', 'bbox': [7, 10, 21, 20]},
                   {'content': 'ZKA', 'bbox': [6, 40, 62, 58]}, ...],}
        """

        # --- find the topest and leftest text in the panel bbox ---
        # the topest and leftest text in the panel bbox
        def is_same_line(bbox1, bbox2, y_threshold):
            _, y1_min, _, y1_max = bbox1
            _, y2_min, _, y2_max = bbox2
            # Calculate the vertical midpoints of the bounding boxes
            y1_mid = (y1_min + y1_max) / 2
            y2_mid = (y2_min + y2_max) / 2
            # Check if the vertical midpoints are close enough to consider them in the same line
            return abs(y1_mid - y2_mid) <= y_threshold

        top_left_text = None
        y_threshold = 5  # Adjust as needed
        if self.software_name == "adobe acrobat":
            if "关闭" not in [t["content"] for t in ocr_result]:
                return True, "Document"
            top_threshold = 80  # Adjust as needed
        else:
            top_threshold = 0
        for text in ocr_result:
            # Check if the text is in the panel bounding box
            if is_in_bbox(text["bbox"], panel_bbox):
                # Initialize top_left_text if it is None
                # print("selected text:", text)
                if top_left_text is None:
                    top_left_text = text
                else:
                    # Check if the text is on the same line as the current top_left_text
                    if is_same_line(text["bbox"], top_left_text["bbox"], y_threshold):
                        # If the text is more towards the left, update top_left_text
                        if text["bbox"][0] < top_left_text["bbox"][0]:
                            top_left_text = text
                    else:
                        # print("selected text:", text, top_left_text)
                        # Check if the text is more towards the top
                        if text["bbox"][1] < top_left_text["bbox"][1] - top_threshold:
                            top_left_text = text
                # print("top_left_text: ", top_left_text)
        # print("top_left_text: ", top_left_text)
        # from IPython.core.debugger import Pdb
        # Pdb().set_trace()
        # the ocr could have noise, postprocess the top_left_text by comparing with predefined panel name
        # TODO: update more panel names here
        print("top_left_text: ", top_left_text)
        if top_left_text is None:
            return False, "Accessory"
        text2panel_name = {
            "效果控件": "Effect Controls",
            "效果 控件": "Effects",
            "节目": "Program",
            "项目": "Project",
            "效果": "Effects",
            "学习": "Learn",
            "基本图形": "Essential Graphics",
            "基本声音": "Essential Sound",
            "Lumetri 颜色": "Lumetri Color",
            "Lumetri 范围": "Lumetri Scopes",
            "标记": "Markers",
            "历史记录": "History",
            "信息": "Info",
            "合成": "Composition",
            "字符": "Character",
            "段落": "Paragraph",
            "跟踪器": "Tracker",
            "预览": "Preview",
            "图层": "Layer",
            "对齐": "Align",
            "编辑 PDF": "Edit PDF",
            "创建 PDF": "Create PDF",
            "合并文件": "Combine Files",
            "导出 PDF": "Export PDF",
            "组织页面": "Organize Pages",
            "填写和签名": "Fill and Sign",
            "注释": "Comment",
            "扫描和OCR": "Scan and OCR",
            "Document": "Document",
            "Effect Controls": "Effect Controls",
            "Effects": "Effects",
            "Effects & Presets": "Effects & Presets",
            "Program": "Program",
            "Project": "Project",
            "Learn": "Learn",
            "Essential Graphics": "Essential Graphics",
            "Essential Sound": "Essential Sound",
            "Lumetri Color": "Lumetri Color",
            "Lumetri Scopes": "Lumetri Scopes",
            "Markers": "Markers",
            "History": "History",
            "Info": "Info",
            "Composition": "Composition",
            "Character": "Character",
            "Paragraph": "Paragraph",
            "Tracker": "Tracker",
            "Preview": "Preview",
            "Layer": "Layer",
            "Audio": "Audio",
            "Align": "Align",
        }
        # some panel name has more information, e.g. '项目: assistgui', so only keep the first part
        if ":" in top_left_text["content"]:
            rough_panel_name = top_left_text["content"].split(":")[0]
        else:
            rough_panel_name = top_left_text["content"]  # .split(' ')[0]
        # if self.timeline_name != None and textdistance.levenshtein(rough_panel_name, self.timeline_name) < 2:
        #     return True, "Timeline"
        # compute the editing distance
        min_distance = 100
        min_distance_panel_name = None
        for candidate_panel_name in text2panel_name.keys():
            distance = textdistance.levenshtein(rough_panel_name, candidate_panel_name)
            if distance < min_distance:
                min_distance = distance
                min_distance_panel_name = candidate_panel_name
        # if the distance is too large, cannot recognize the panel with OCR
        if min_distance >= 2:
            return False, "Accessory"
        else:
            if text2panel_name[min_distance_panel_name] == "Program":
                self.timeline_name = top_left_text["content"]
            return True, text2panel_name[min_distance_panel_name]

    def recognize_panel_with_icon(self, panel_bbox, screenshot_path):
        # TODO: check if some icons are in the panel
        success = False
        panel_name = None
        panel_img = crop_panel(panel_bbox, screenshot_path)
        button_box = detect_button(
            panel_img, software_name=self.software_name, panel_name="Tools"
        )
        if len(button_box) > 0:
            success = True
            panel_name = "Tools"
        return success, panel_name

    def get_software_name(self):
        return self.software_name



import os
import ast
import base64
from io import BytesIO
from pathlib import Path
from uuid import uuid4
import datetime

import platform
import pyautogui
import requests
import torch
from PIL import Image, ImageDraw
from qwen_vl_utils import process_vision_info
from transformers import AutoProcessor, Qwen2VLForConditionalGeneration

from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import encode_image
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_showui, colorful_text_vlm

from transformers import AutoModelForCausalLM, AutoTokenizer

from ..llm_utils.run_litellm import run_litellm
import json
os.environ["TOKENIZERS_PARALLELISM"] = "false"


class LLMActor:
    def __init__(self, llm_model, output_callback, device=torch.device("cpu"), local=True):
        self.device = device
        # self.split = split
        self.output_callback = output_callback
        self.os_name = platform.system()

        if llm_model in ["qwen2.5"]:
            # load local model
            model_name = "Qwen/Qwen2.5-7B-Instruct"
            self.model = AutoModelForCausalLM.from_pretrained(
                model_name,
                torch_dtype="auto",
                device_map="auto"
            )
            self.tokenizer = AutoTokenizer.from_pretrained(model_name)

        elif llm_model in ['gpt-4o', 'gpt-4o-mini']:
            # do nothing
            self.model = llm_model

        self.action_history = ''  # Initialize action history

    def __call__(self, messages, parsed_gui, screenshot_path):

        task = messages
        formatted_gui = self.format_parsed_gui(parsed_gui)
        prompt_message = self.get_prompt(formatted_gui, task)
        system_prompt = self.get_system_prompt()

        # Call llm to get the next action
        response = run_litellm(
            messages=prompt_message, 
            system=system_prompt, 
            llm=self.model
        )

        # TODO: uncomment this for support gradio image display
        image_base64 = encode_image(screenshot_path)
        # self.output_callback(f'Screenshot for {colorful_text_showui}:\n<img src="data:image/png;base64,{image_base64}">', sender="bot")
        print("response:", response)
        # Update action history
        result = response[0]
        
        self.action_history += result + '\n'
        # 从response中提取json内容

        
        # Return response in expected format
        response = {'content': result, 'role': 'assistant'}
        return response

    def parse_showui_output(self, output_text):
        try:
            # Ensure the output is stripped of any extra spaces
            output_text = output_text.strip()

            # Wrap the input in brackets if it looks like a single dictionary
            if output_text.startswith("{") and output_text.endswith("}"):
                output_text = f"[{output_text}]"

            # Validate if the output resembles a list of dictionaries
            if not (output_text.startswith("[") and output_text.endswith("]")):
                raise ValueError("Output does not look like a valid list or dictionary.")

            # Parse the output using ast.literal_eval
            parsed_output = ast.literal_eval(output_text)

            # Ensure the result is a list
            if isinstance(parsed_output, dict):
                parsed_output = [parsed_output]
            elif not isinstance(parsed_output, list):
                raise ValueError("Parsed output is neither a dictionary nor a list.")

            # Ensure all elements in the list are dictionaries
            if not all(isinstance(item, dict) for item in parsed_output):
                raise ValueError("Not all items in the parsed output are dictionaries.")

            return parsed_output

        except Exception as e:
            print(f"Error parsing output: {e}")
            return None

    
    @staticmethod
    def format_parsed_gui(gui_data):
        result = []
        # Loop through the 'teach_mode_GUI' list
        for section_name, section in gui_data.items():
            for panel in section:
                elements = panel.get('elements', [[]])
                for group in elements:
                    row_result = []  # Collect items row-wise
                    for item in group:
                        name = item.get('name', '')
                        rectangle = item.get('rectangle', [])
                        # Calculate the midpoint of the bounding box
                        x1, y1, x2, y2 = map(int, rectangle)
                        mid_x = (x1 + x2) // 2
                        mid_y = (y1 + y2) // 2
                        row_result.append(f"{name} [{mid_x}, {mid_y}]")
                    result.append('; '.join(row_result))  # Join row items with semicolon
        
        # Join all formatted rows into a single string
        return '\n'.join(result)

    def get_prompt(self, formatted_gui, task):
        return f"""The information shown in the screenshot is:
{formatted_gui}


What you need to do is:
{task}

Output the next action in the following format:
```json
{{"action": "ACTION_TYPE", "text": "text_to_type", "coordinate": [x,y]}}
```

Let's do it:
"""

    def get_system_prompt(self):
        DATETIME = datetime.datetime.now().strftime("%A, %B %d, %Y")
        return f"""System Overview:
You have access to a set of functions that allow you to interact with a computing environment. 

String and scalar parameters should be passed as is. Lists and objects should be passed in JSON format.
You can respond to the user based on the results or make further function calls.

Available Functions:

Computer Interaction (GUI):
- Description: 
    Use a mouse and keyboard to interact with the computer and take screenshots.
    You can only interact with the desktop GUI (no terminal or application menu access).
- Actions include:
    - key: Press a key or key-combination.
    - type: Type a string of text.
    - mouse_move: Move the cursor to specified coordinates.
    - left_click, right_click, middle_click, double_click: Perform mouse clicks.
    - left_click_drag: Click and drag the cursor.
    - get_statistics_info: Get the statistical information when it is shown in the screenshot and user wants to know.
- Parameters:
    - action (required): The action to perform, such as key, type, mouse_move, etc.
    - coordinate: The (x, y) coordinates for mouse-related actions.
    - text: The text to type or key to press for type and key actions.

System Capabilities:
You are using an {self.os_name} system.
The current date is [{DATETIME}].
"""




if __name__ == "__main__":
    # Initialize the actor
    actor = ShowUIActor(
        model_path="",
        device=torch.device("mps"),
        split='web',
    )

    # Define the task
    task = "Search for 'weather for New York city'."

    # Initialize messages
    messages = [{"role": "user", "content": task}]

    # Simulate multiple interactions
    for _ in range(3):  # Adjust the range as needed
        # Call the actor
        response = actor(messages)
        # Print the response
        print("Response from actor:", response)
        # Add assistant's response to messages
        messages.append(response)



"""
Agentic sampling loop that calls the Anthropic API and local implementation of anthropic-defined computer use tools.
"""
import asyncio
import platform
from collections.abc import Callable
from datetime import datetime
from enum import StrEnum
from typing import Any, cast

from anthropic import Anthropic, AnthropicBedrock, AnthropicVertex, APIResponse
from anthropic.types import (
    ToolResultBlockParam,
)
from anthropic.types.beta import (
    BetaContentBlock,
    BetaContentBlockParam,
    BetaImageBlockParam,
    BetaMessage,
    BetaMessageParam,
    BetaTextBlockParam,
    BetaToolResultBlockParam,
)
from anthropic.types import TextBlock
from anthropic.types.beta import BetaMessage, BetaTextBlock, BetaToolUseBlock

from computer_use_ootb_internal.computer_use_demo.tools import BashTool, ComputerTool, EditTool, ToolCollection, ToolResult

from PIL import Image
from io import BytesIO
import gradio as gr
from typing import Dict


BETA_FLAG = "computer-use-2024-10-22"


class APIProvider(StrEnum):
    ANTHROPIC = "anthropic"
    BEDROCK = "bedrock"
    VERTEX = "vertex"
    OPENAI = "openai"
    QWEN = "qwen"


PROVIDER_TO_DEFAULT_MODEL_NAME: dict[APIProvider, str] = {
    APIProvider.ANTHROPIC: "claude-3-5-sonnet-20241022",
    APIProvider.BEDROCK: "anthropic.claude-3-5-sonnet-20241022-v2:0",
    APIProvider.VERTEX: "claude-3-5-sonnet-v2@20241022",
}


# Check OS
SYSTEM_PROMPT = f"""<SYSTEM_CAPABILITY>
* You are utilizing a Windows system with internet access.
* The current date is {datetime.today().strftime('%A, %B %d, %Y')}.
</SYSTEM_CAPABILITY>
"""


class AnthropicActor:
    def __init__(
        self, 
        model: str, 
        provider: APIProvider, 
        system_prompt_suffix: str, 
        api_key: str,
        api_response_callback: Callable[[APIResponse[BetaMessage]], None],
        max_tokens: int = 4096,
        only_n_most_recent_images: int | None = None,
        selected_screen: int = 0,
        print_usage: bool = True,
    ):
        self.model = model
        self.provider = provider
        self.system_prompt_suffix = system_prompt_suffix
        self.api_key = api_key
        self.api_response_callback = api_response_callback
        self.max_tokens = max_tokens
        self.only_n_most_recent_images = only_n_most_recent_images
        self.selected_screen = selected_screen
        
        self.tool_collection = ToolCollection(
            ComputerTool(selected_screen=selected_screen),
            BashTool(),
            EditTool(),
        )

        self.system = (
            f"{SYSTEM_PROMPT}{' ' + system_prompt_suffix if system_prompt_suffix else ''}"
        )
        
        self.total_token_usage = 0
        self.total_cost = 0
        self.print_usage = print_usage

        # Instantiate the appropriate API client based on the provider
        if provider == APIProvider.ANTHROPIC:
            self.client = Anthropic(api_key=api_key)
        elif provider == APIProvider.VERTEX:
            self.client = AnthropicVertex()
        elif provider == APIProvider.BEDROCK:
            self.client = AnthropicBedrock()

    def __call__(
        self, 
        *,
        messages: list[BetaMessageParam]
    ):
        """
        Generate a response given history messages.
        """
        if self.only_n_most_recent_images:
            _maybe_filter_to_n_most_recent_images(messages, self.only_n_most_recent_images)

        # Call the API synchronously
        raw_response = self.client.beta.messages.with_raw_response.create(
            max_tokens=self.max_tokens,
            messages=messages,
            model=self.model,
            system=self.system,
            tools=self.tool_collection.to_params(),
            betas=["computer-use-2024-10-22"],
        )

        self.api_response_callback(cast(APIResponse[BetaMessage], raw_response))

        response = raw_response.parse()
        print(f"AnthropicActor response: {response}")

        self.total_token_usage += response.usage.input_tokens + response.usage.output_tokens
        self.total_cost += (response.usage.input_tokens * 3 / 1000000 + response.usage.output_tokens * 15 / 1000000)
        
        if self.print_usage:
            print(f"Claude total token usage so far: {self.total_token_usage}, total cost so far: $USD{self.total_cost}")
        
        return response


def _maybe_filter_to_n_most_recent_images(
    messages: list[BetaMessageParam],
    images_to_keep: int,
    min_removal_threshold: int = 10,
):
    """
    With the assumption that images are screenshots that are of diminishing value as
    the conversation progresses, remove all but the final `images_to_keep` tool_result
    images in place, with a chunk of min_removal_threshold to reduce the amount we
    break the implicit prompt cache.
    """
    if images_to_keep is None:
        return messages

    tool_result_blocks = cast(
        list[ToolResultBlockParam],
        [
            item
            for message in messages
            for item in (
                message["content"] if isinstance(message["content"], list) else []
            )
            if isinstance(item, dict) and item.get("type") == "tool_result"
        ],
    )

    total_images = sum(
        1
        for tool_result in tool_result_blocks
        for content in tool_result.get("content", [])
        if isinstance(content, dict) and content.get("type") == "image"
    )

    images_to_remove = total_images - images_to_keep
    # for better cache behavior, we want to remove in chunks
    images_to_remove -= images_to_remove % min_removal_threshold

    for tool_result in tool_result_blocks:
        if isinstance(tool_result.get("content"), list):
            new_content = []
            for content in tool_result.get("content", []):
                if isinstance(content, dict) and content.get("type") == "image":
                    if images_to_remove > 0:
                        images_to_remove -= 1
                        continue
                new_content.append(content)
            tool_result["content"] = new_content
            
            

if __name__ == "__main__":
    pass
    # client = Anthropic(api_key="")
    # response = client.beta.messages.with_raw_response.create(
    #     max_tokens=4096,
    #     model="claude-3-5-sonnet-20241022",
    #     system=SYSTEM_PROMPT,
    #     # tools=ToolCollection(
    #     #     ComputerTool(selected_screen=0),
    #     #     BashTool(),
    #     #     EditTool(),
    #     # ).to_params(),
    #     betas=["computer-use-2024-10-22"],
    #     messages=[
    #         {"role": "user", "content": "click on (199, 199)."}
    #     ],
    # )
    
    # print(f"AnthropicActor response: {response.parse().usage.input_tokens+response.parse().usage.output_tokens}")


# -*- coding: utf-8 -*-
# Please pull the latest code to run the profiling.
import time
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers.generation import GenerationConfig
from transformers.trainer_utils import set_seed
from tqdm import tqdm
import os

seed = 1024
max_experiment_times = 1
context_length_per_experiment = 258
generate_length_per_experiment = 2048 - 258
# context_length_per_experiment = 258
# generate_length_per_experiment = 8192 - 258
# context_length_per_experiment = 2048
# generate_length_per_experiment = 1
use_flash_attn = True
# fp32 without quantization, bf16 without quantization or int4 quantization (updated on 8.28: we have updated the quantization strategy to AutoGPTQ in this script)
quant_type = "bf16" # fp32, bf16 or int4

set_seed(seed)

tokenizer = AutoTokenizer.from_pretrained("Qwen/Qwen-VL-Chat", trust_remote_code=True)

if quant_type == "bf16":
    model = AutoModelForCausalLM.from_pretrained(
        "Qwen/Qwen-VL-Chat", 
        device_map="mps", 
        # device_map="cuda:0", 
        trust_remote_code=True, 
        bf16=True, 
        use_flash_attn=use_flash_attn
    ).eval()
elif quant_type == "fp32":
    assert use_flash_attn is False, "FP32 profiling cannot be performed with Flash-Attention activated."
    model = AutoModelForCausalLM.from_pretrained(
        "Qwen/Qwen-VL-Chat", 
        # device_map="cuda:0", 
        device_map="mps", 
        trust_remote_code=True, 
        fp32=True, 
        use_flash_attn=False
    ).eval()    
elif quant_type == "int4":
    # please install AutoGPTQ following the readme to use quantization
    from auto_gptq import AutoGPTQForCausalLM
    model = AutoGPTQForCausalLM.from_quantized(
        "Qwen/Qwen-VL-Chat-Int4", 
        # device="cuda:0", 
        device="mps", 
        trust_remote_code=True, 
        use_safetensors=True, 
        use_flash_attn=use_flash_attn
    ).eval()

# Specify hyperparameters for generation
config = GenerationConfig.from_pretrained("Qwen/Qwen-VL-Chat", trust_remote_code=True)
config.min_length = generate_length_per_experiment + context_length_per_experiment
config.max_new_tokens = generate_length_per_experiment

time_costs = []

# TODO: download an image
if not os.path.exists("./clothes1.jpeg"):
    import requests
    url = "XXX"
    r = requests.get(url)
    with open("./clothes1.jpeg", "rb") as f:
        f.write(r.content)
        f.close()

context_str = '<img>./clothes1.jpeg</img>' + '我' * max(0, context_length_per_experiment - 258)
max_gpu_memory_cost = 0
for _ in tqdm(range(max_experiment_times)):
    inputs = tokenizer(context_str, return_tensors='pt')
    inputs = inputs.to(model.device)
    t1 = time.time()
    pred = model.generate(**inputs, generation_config=config)
    time_costs.append(time.time() - t1)
    assert pred.shape[1] == config.min_length
    max_gpu_memory_cost = max(max_gpu_memory_cost, torch.cuda.max_memory_allocated())
    torch.cuda.empty_cache()

print("Average generate speed (tokens/s): {}".format((max_experiment_times * generate_length_per_experiment) / sum(time_costs)))
print(f"GPU Memory cost: {max_gpu_memory_cost / 1024 / 1024 / 1024}GB")
print("Experiment setting: ")
print(f"seed = {seed}")
print(f"max_experiment_times = {max_experiment_times}")
print(f"context_length_per_experiment = {context_length_per_experiment}")
print(f"generate_length_per_experiment = {generate_length_per_experiment}")
print(f"use_flash_attn = {use_flash_attn}")
print(f"quant_type = {quant_type}")



import os
import logging
import base64
import requests

import dashscope
# from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils import is_image_path, encode_image

def is_image_path(text):
    return False

def encode_image(image_path):
    return ""   


def run_qwen(messages: list, system: str, llm: str, api_key: str, max_tokens=256, temperature=0):
    
    api_key = api_key or os.environ.get("QWEN_API_KEY")
    if not api_key:
        raise ValueError("QWEN_API_KEY is not set")
    
    dashscope.api_key = api_key
    
    # from IPython.core.debugger import Pdb; Pdb().set_trace()

    final_messages = [{"role": "system", "content": [{"text": system}]}]
    # image_url = "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
    if type(messages) == list:
        for item in messages:
            contents = []
            if isinstance(item, dict):
                for cnt in item["content"]:
                    if isinstance(cnt, str):
                        if is_image_path(cnt):
                            # base64_image = encode_image(cnt)
                            content = [{"image": cnt}]
                        # content = {"type": "image_url", "image_url": {"url": image_url}}
                    else:
                        content = {"text": cnt}
                    contents.append(content)
                    
                message = {"role": item["role"], "content": contents}
            else:  # str
                contents.append({"text": item})
                message = {"role": "user", "content": contents}
            
            final_messages.append(message)

    print("[qwen-vl] sending messages:", final_messages)

    response = dashscope.MultiModalConversation.call(
        model='qwen-vl-max-latest',
        # model='qwen-vl-max-0809',
        messages=final_messages
        )

    # from IPython.core.debugger import Pdb; Pdb().set_trace()
    
    try:
        text = response.output.choices[0].message.content[0]['text']
        usage = response.usage
        
        if "total_tokens" not in usage:
            token_usage = int(usage["input_tokens"] + usage["output_tokens"])
        else:
            token_usage = int(usage["total_tokens"])
        
        return text, token_usage
        # return response.json()['choices'][0]['message']['content']
    # return error message if the response is not successful
    except Exception as e:
        print(f"Error in interleaved openAI: {e}. This may due to your invalid OPENAI_API_KEY. Please check the response: {response.json()} ")
        return response.json()



if __name__ == "__main__":
    api_key = os.environ.get("QWEN_API_KEY")
    if not api_key:
        raise ValueError("QWEN_API_KEY is not set")
    
    dashscope.api_key = api_key
    
    final_messages = [{"role": "user",
                       "content": [
                           {"text": "What is in the screenshot?"},
                           {"image": "./tmp/outputs/screenshot_0b04acbb783d4706bc93873d17ba8c05.png"}
                           ]
                       }
                    ]
    response = dashscope.MultiModalConversation.call(model='qwen-vl-max-0809', messages=final_messages)
    
    print(response)
    
    text = response.output.choices[0].message.content[0]['text']
    usage = response.usage
    
    if "total_tokens" not in usage:
        if "image_tokens" in usage:
            token_usage = usage["input_tokens"] + usage["output_tokens"] + usage["image_tokens"]
        else:
            token_usage = usage["input_tokens"] + usage["output_tokens"]
    else:
        token_usage = usage["total_tokens"]
    
    print(text, token_usage)
    # The screenshot is from a video game... 1387